""" This module implememts the class `RepositoryManager` which manages 
    all the access to the project repository.
    An important use of the class is to avoid races during 
    parallel processing. This also implements methods to keep track
    of the changes made to the repository during executions: 
    (the repository is backed before any execution and can be restored after)

    Each method that manipulate the repository, except backing up and
    restoring, take one or two optional callback functions and 
    callback arguments as last arguments 
    (callback function mandatory only for custom_read_access), 
    which will be called after locking the repository, 
    with the last parameters:
        - repository root dir 
        - executable relative path w.r.t the repository root dir
        - list of source files of interest 
                            (as relative path from repository root dir)
        - list of developer test cases of interest

    Callbacks signature:
    :param op_retval: value return by the operation before the callback. 
        None is the value if no operation occured 
        (ex: `RepositorManager.custom_access`)
        Note: this is only present for post processing callbacks
    :param callback_args: data object as understandable by the callback.
        The same data is passed to the method of `RepositoryManager` that 
        call the call back. it is then passed to the 
        callback without any change.
    :param str repository_rootdir: full path to the repository root
    :param str repo_executables_relpaths: relative path to the executable
        w.r.t the repository root dir
    :param list source_files_list: list of source files we care about.
        each item of the list is a relative path to a source file w.r.t. 
        the repository root dir
    :param list dev_tests_list: list of developer tests we care about.
        each item of the list is a relative path to a test file w.r.t. 
        the repository root dir. 
    :return: The returned value is the returned value of the operation 
        before the call back (in case the callback function is None), or
        the returned value of the callback function.
    >>> def post_process_callback(op_retval, callback_args,\
                                    repository_rootdir, \
                                    repo_executables_relpaths, \
                                    source_files_list, \
                                    dev_tests_list)
    
    TODO: IMPORTANT BEFORE ANY PARALLELISM: make sure to avoid deadlock
        and release the lock when there is failure (call to error_exit).
        Implement tools with subprocess for parallelism so as to kill all the 
        subprocess upon error_exit, or continue until join.
"""


from __future__ import print_function

import os
import shutil
import logging
import threading

# https://gitpython.readthedocs.io/en/stable/
from git import Repo as git_repo
import git.exc as git_exc
logging.getLogger("git.cmd").setLevel(logging.INFO)

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

ERROR_HANDLER = common_mix.ErrorHandler

class RepositoryManager(object):
    DEFAULT_TESTS_BRANCH_NAME = "_tests_tmp_muteria_"
    DEFAULT_MUTERIA_REPO_META_FOLDER = ".muteria" 
    def __init__(self, repository_rootdir, repo_executables_relpaths=None, \
                        dev_test_runner_func=None, code_builder_func=None, \
                        dev_test_program_wrapper=None, \
                        test_exec_output_cleaner_func=None, \
                        source_files_to_objects=None, dev_tests_list=None, \
                        delete_created_on_revert_as_initial=False, \
                        test_branch_name=DEFAULT_TESTS_BRANCH_NAME):
        self.repository_rootdir = repository_rootdir
        self.repo_executables_relpaths = repo_executables_relpaths
        self.dev_test_runner_func = dev_test_runner_func
        self.code_builder_func = code_builder_func
        self.source_files_to_objects = source_files_to_objects
        self.dev_tests_list = dev_tests_list
        self.dev_test_program_wrapper = dev_test_program_wrapper
        if self.dev_test_program_wrapper is not None:
            self.dev_test_program_wrapper = self.dev_test_program_wrapper(self)
        self.test_exec_output_cleaner_func = test_exec_output_cleaner_func
        if self.test_exec_output_cleaner_func is None:
            self.test_exec_output_cleaner_func = lambda x: x

        self.delete_created_on_revert_as_initial = \
                                        delete_created_on_revert_as_initial
        self.test_branch_name = test_branch_name

        if self.repo_executables_relpaths is None:
            self.repo_executables_relpaths = []
        if self.dev_tests_list is None:
            self.dev_tests_list = []
        if self.source_files_to_objects is None:
            self.source_files_to_objects = {}

        self.source_files_list = list(self.source_files_to_objects.keys())

        if self.repository_rootdir is None:
            ERROR_HANDLER.error_exit(\
                                "repository rootdir cannot be None", __file__)
        #if self.repo_executables_relpaths is None:
        #    ERROR_HANDLER.error_exit(\
        #                    "repo executable relpath cannot be None", __file__)

        # TODO: Implement a mechanism to avoid deadlock (multiple levels of
        # parallelism)
        self.lock = threading.RLock()

        self.muteria_metadir = os.path.join(self.repository_rootdir, \
                                        self.DEFAULT_MUTERIA_REPO_META_FOLDER)
        self.muteria_metadir_info_file = os.path.join(self.muteria_metadir, \
                                                            "src_infos.json")

        # setup the repo (Should remain as last intruction of initialization)
        self._setup_repository()
    #~ def __init__()

    def get_dev_tests_list(self):
        return self.dev_tests_list
    #~ def get_dev_tests_list()

    def has_dev_tests(self):
        return (self.dev_test_runner_func is not None and \
                                            self.dev_tests_list is not None)
    #~ def has_dev_tests()

    def get_wrapper_object(self):
        return self.dev_test_program_wrapper
    #~ def get_wrapper_object()

    def get_test_exec_output_cleaner_func (self):
        return self.test_exec_output_cleaner_func 
    #~ def get_test_exec_output_cleaner_func()

    def should_build(self):
        return (self.code_builder_func is not None)
    #~ def should_build()

    def _set_callback_basics(self, callback_object):
        if callback_object is not None:
            callback_object.set_repository_rootdir(self.repository_rootdir)
            callback_object.set_repo_executables_relpaths(\
                                             self.repo_executables_relpaths)
            callback_object.set_source_files_to_objects(\
                                                self.source_files_to_objects)
            callback_object.set_dev_tests_list(self.dev_tests_list)
    #~ def _set_callback_basics()

    def run_dev_test(self, dev_test_name, exe_path_map=None, env_vars=None,\
                    timeout=None, collected_output=None, callback_object=None):
        if self.dev_test_runner_func is None:
            ERROR_HANDLER.error_exit(\
                    "dev_test_runner_func cannot be none when called", \
                                                                    __file__)

        pre_ret = common_mix.GlobalConstants.COMMAND_SUCCESS
        post_ret = common_mix.GlobalConstants.COMMAND_UNCERTAIN
        ret = common_mix.GlobalConstants.COMMAND_UNCERTAIN

        self._set_callback_basics(callback_object)

        self.lock.acquire()
        try:
            if callback_object is not None:
                pre_ret = callback_object.before_command()
            if pre_ret == common_mix.GlobalConstants.COMMAND_SUCCESS:
                ret = self.dev_test_runner_func(dev_test_name, \
                                        self.repository_rootdir, \
                                        exe_path_map=exe_path_map, \
                                        env_vars=(env_vars 
                                                    if env_vars is not None \
                                                    else {}), \
                                        timeout=timeout, \
                                        collected_output=collected_output, \
                                        using_wrapper=(\
                                                    self.get_wrapper_object() \
                                                                is not None))
                                        #self.repo_executables_relpaths)
                if callback_object is not None:
                    callback_object.set_op_retval(ret)
                    post_ret = callback_object.after_command()
        finally:
            self.lock.release()                                
        return (pre_ret, ret, post_ret)
    #~ def run_dev_test()

    def build_code(self, compiler=None, flags_list=None, clean_tmp=False, \
                        reconfigure=False, callback_object=None):
        """ Build the code in repository dir to obtain the executable
        
        :type compiler: str
        :param compiler: name of the compiler to usei. default to None.

        :type flags_list: list
        :param flags_list: list of strings to use as compiler flags. 
                        default to None.

        :type clean_tmp: bool
        :param clean_tmp: enable clean any build temporary files, 
                such as object files in C/C++ . Default to False.

        :type \reconfigure: bool
        :param \reconfigure: enable reconfigure the project repository before
                building. may be usefule if using a different compiler
                on different runs. default to False.
        :type pre_process_callback: funtion
        :param pre_process_callback: function that will be called before
                the code is build. 
                the first argument is the `pre_callback_args` passed 
                to the function. the remaining four arguments are the
                project element (repos rootdir absolute path, 
                executable relative path, list of source files,
                list of developer tests). Default to None.
                The build script and `post_process_callback` are called
                if and only if this function return True.

        :type pre_callback_args: anything
        :param pre_callback_args: data that will be passed (untouched) to the
                callback function `pre_process_callback`. Default to None.

        :type post_process_callback: funtion
        :param post_process_callback: function that will be called after
                the code is build. The call pass first the returned
                value of the build (True is succeed, False if failed).
                the second argument is the `callback_args` passed 
                to the function. the remaining four arguments are the
                project element (repos rootdir absolute path, 
                executable relative path, list of source files,
                list of developer tests). Default to None.

        :type post_callback_args: anything
        :param post_callback_args: data that will be passed (untouched) to the
                callback function `post_process_callback`. Default to None.

        :raises: Terminates execution with error if the specified
                `self.code_builder_func` is None.

        :rtype: The returned value is the pair of returned value of 
                `pre_process_callback` and `post_process_callback`.
                if `pre_process_callback` is None, the default return is True.
                if `post_process_callback` is None, the default return is None.
        """

        if self.code_builder_func is None:
            ERROR_HANDLER.error_exit(\
                    "code_builder_func cannot be none when called", __file__)

        pre_ret = common_mix.GlobalConstants.COMMAND_SUCCESS
        post_ret = common_mix.GlobalConstants.COMMAND_UNCERTAIN
        ret = common_mix.GlobalConstants.COMMAND_UNCERTAIN

        self._set_callback_basics(callback_object)

        self.lock.acquire()
        try:
            if callback_object is not None:
                pre_ret = callback_object.before_command()
            if pre_ret == common_mix.GlobalConstants.COMMAND_SUCCESS:
                ret = self.code_builder_func(self.repository_rootdir, \
                                        self.repo_executables_relpaths, \
                                        compiler, flags_list, clean_tmp, \
                                        reconfigure)
                if callback_object is not None:
                    callback_object.set_op_retval(ret)
                    post_ret = callback_object.after_command()
        finally:
            self.lock.release()                                
        return (pre_ret, ret, post_ret)
    #~ def build_code()

    def custom_read_access(self, callback_object):
        """ Execute the read only callback function and 
            return its returned value
        :type process_callback: function
        :param process_callback: function that will be called after
                the code is build. The call pass first the returned
                value of the build (True is succeed, False if failed).
                the second argument is the `callback_args` passed 
                to the function. the remaining four arguments are the
                project element (repos rootdir absolute path, 
                executable relative path, list of source files,
                list of developer tests). Default to None.

        :type callback_args: anything
        :param callback_args: data that will be passed (untouched) to the
                callback function `process_callback`. Default to None.

        :raises: error if process_callback is None

        :rtype: any
        """

        self.revert_src_list_files()
        
        self._set_callback_basics(callback_object)

        self.lock.acquire()
        try:
            if callback_object is None:
                ERROR_HANDLER.error_exit("{} {}".format(\
                        "callback object must", \
                        "not be None in custom_read_access call"), __file__)
            post_ret = common_mix.GlobalConstants.COMMAND_UNCERTAIN
            pre_ret = callback_object.before_command()
            if pre_ret == common_mix.GlobalConstants.COMMAND_SUCCESS:
                callback_object.set_op_retval(pre_ret)
                post_ret = callback_object.after_command()
        finally:
            self.lock.release()                                
        return (pre_ret, post_ret)
    #~ def custom_read_access()

    def get_repository_dir_path(self):
        return self.repository_rootdir
    #~ def get_repository_dir_path()

    def repo_abs_path(self, relpath):
        return os.path.join(self.repository_rootdir, relpath)
    #~ def repo_abs_path()

    def get_relative_exe_path_map(self):
        exes = list(self.repo_executables_relpaths)
        src_obj_map = dict(self.source_files_to_objects)
        return exes, src_obj_map
    #~ def get_relative_exe_path_map()

    def get_exe_path_map(self):
        exes = [self.repo_abs_path(e) for e in self.repo_executables_relpaths]
        src_obj_map = {}
        for src, dest in list(self.source_files_to_objects.items()):
            src_obj_map[self.repo_abs_path(src)] = \
                            dest if dest is None else self.repo_abs_path(dest)
        return exes, src_obj_map
    #~ def get_exe_path_map()

    def _unlocked_revert_repository_file (self, file_rel_path, gitobj=None):
        if gitobj is None:
            repo = git_repo(self.repository_rootdir)
            gitobj = repo.git
        try:    
            gitobj.checkout('--', file_rel_path)
        except OSError as e:
            ERROR_HANDLER.error_exit("git checkout command " + \
                                    "(-- {}) failed ".format(file_rel_path) + \
                                    "with error: {}".format(str(e)), __file__)
    #~ def _unlocked_revert_repository_file()

    def revert_repository_file (self, file_rel_path, gitobj=None):
        self.lock.acquire()
        try:
            self._unlocked_revert_repository_file(file_rel_path, gitobj)
        finally:
            self.lock.release()                                
    #~ def revert_repository_file()

    def revert_src_list_files (self):
        self.lock.acquire()
        try:
            repo = git_repo(self.repository_rootdir)
            gitobj = repo.git
            #for src in self.source_files_list:
            #    self._unlocked_revert_repository_file(src, gitobj=gitobj)
            self._unlocked_revert_repository_file(self.source_files_list, \
                                                                gitobj=gitobj)
        finally:
            self.lock.release()                                
    #~ def revert_src_list_files()

    def revert_repository(self, as_initial=False):
        #repo = git_repo(self.repository_rootdir)
        #gitobj = repo.git
        self.lock.acquire()
        try:
            if as_initial:
                if self.delete_created_on_revert_as_initial:
                    self._delete_testing_branch(self.repository_rootdir, \
                                                self.test_branch_name)
                else:
                    self.lock.release()                                
                    ERROR_HANDLER.error_exit("{} {}".format(\
                            "revert_repository called with", \
                            "'delete_created_on_revert_as_initial' disabled"),\
                                                                    __file__)
            else:
                # Reset the files but do not delete created files and dir
                self.revert_src_list_files()
                shutil.rmtree(self.muteria_metadir)
                #gitobj.reset('--hard') 
        finally:
            self.lock.release()                                
    #~ def revert_repository()

    def _setup_repository(self):
        # Make sure the repo dir exists
        ERROR_HANDLER.assert_true(os.path.isdir(self.repository_rootdir), \
                        "given repository dir is not existing: {}". format( \
                            self.repository_rootdir), __file__)
        
        # make sure the repo dir is a git repo
        # if no, ask the user whether to initialize and initialize or abort
        # if yes or user accepted initialize, get the git object for the repo
        try:
            repo = git_repo(self.repository_rootdir)
            gitobj = repo.git
        except git_exc.InvalidGitRepositoryError:
            make_it_git = common_mix.confirm_execution("{} {} {} {}".format(\
                                    "The given repository directory is not", \
                                    "a git repository, this must be a git", \
                                    "repository to proceed.\n Do you want to",\
                                    "set is as git repository?"))
            if make_it_git:
                repo = git_repo.init(self.repository_rootdir)
                gitobj = repo.git
            else:
                ERROR_HANDLER.error_exit("{} {}".format(\
                            "Must make the repository as git repository,", \
                            "then re-run"), __file__)

        # Check whether the repo is already managed by muteria
        ## if using branch
        if self.delete_created_on_revert_as_initial:
            if self.test_branch_name not in self._get_branches_list(\
                                                self.repository_rootdir):
                ## Not managed, create branch
                self._make_testing_branch(self.repository_rootdir, \
                                                        self.test_branch_name)

            # checkout
            gitobj.checkout(self.test_branch_name)
        
        # Actual Check whether the repo is already managed by muteria
        # There must be a directory DEFAULT_MUTERIA_REPO_META_FOLDER  
        src_in_prev = set()
        if os.path.isdir(self.muteria_metadir):
            ## Managed
            prev_src_list, prev_use_branch = \
                            common_fs.loadJSON(self.muteria_metadir_info_file)
            ERROR_HANDLER.assert_true(prev_use_branch == \
                        self.delete_created_on_revert_as_initial, \
                        "{} {} {} {} {}".format(\
                            "unmatching repo backup type.", \
                            "previously, use branch was", \
                            prev_use_branch, ", now it is",\
                            self.delete_created_on_revert_as_initial), \
                                                                    __file__)
            src_in_prev = set(prev_src_list) & set(self.source_files_list)
            prev_src_list = set(prev_src_list) - set(self.source_files_list)
            remain_prev_src_set = {src for src in prev_src_list \
                                    if os.path.isfile(self.repo_abs_path(src))}
            # make sure that all prev_src are in initial state
            untracked_diff_files = self._get_untracked_and_diffed(repo)
            prev_untracked_diff = remain_prev_src_set & untracked_diff_files
            if len(prev_untracked_diff) > 0:
                bypass = common_mix.confirm_execution(\
                            "{} {} {} {} {}".format(
                                "the following files were previously used as",\
                                "src files by muteria and are now untracked:",\
                                prev_untracked_diff, \
                                "\nDo you want to handle it or bypass it?",\
                                "Choose yes to bypass: "))
                if not bypass:
                    revert_them = common_mix.confirm_execution(\
                            "{} {}".format(\
                                "Do you want to automatically restore the",\
                                "The untracked previous source and continue?"))
                    if revert_them:
                        for src in prev_untracked_diff:
                            self.revert_repository_file(src, gitobj=gitobj)
                    else:
                        ERROR_HANDLER.error_exit(\
                                "Handle it manually and restart the execution")

            # update the info_file
            if set(prev_src_list) != set(self.source_files_list):
                common_fs.dumpJSON([self.source_files_list, \
                                    self.delete_created_on_revert_as_initial],\
                                                self.muteria_metadir_info_file)
        else:
            ## Not managed
            os.mkdir(self.muteria_metadir)
            common_fs.dumpJSON([self.source_files_list, \
                                    self.delete_created_on_revert_as_initial],\
                                                self.muteria_metadir_info_file)

        # Make sure all source files of interest are tracked
        untracked_diff_files = self._get_untracked_and_diffed(repo)
        untracked_diff_src_in_prev = untracked_diff_files & src_in_prev
        if len(untracked_diff_src_in_prev) > 0:
            if common_mix.confirm_execution(\
                            "{} {} {} {} {}".format("The following source",\
                                        "files of interest are untracked", \
                    "and will be reverted (previous execution unfinished):", \
                                        src_in_prev, \
                                        "do you want to revert them?")):
                for src in src_in_prev:
                    self.revert_repository_file(src, gitobj=gitobj)
            else:
                ERROR_HANDLER.error_exit("{} {}".format(\
                                    "Handle untracked source files manually", \
                                                "then restart the execution"))

        untracked_diff_files = self._get_untracked_and_diffed(repo)
        untracked_diff_src_files = set(self.source_files_list) & \
                                                        untracked_diff_files
        if len(untracked_diff_src_files) > 0:
            if common_mix.confirm_execution(\
                                "{} {} {} {}".format("The following source",\
                                        "files of interest are untracked:", \
                                        untracked_diff_src_files, \
                                        "do you want to track them?")):
                repo.index.add(list(untracked_diff_src_files))
            else:
                ERROR_HANDLER.error_exit("{} {}".format(\
                                    "Handle untracked source files manually", \
                                                "then restart the execution"))

    #~ _setup_repository()

    def _get_untracked_and_diffed(self, repo_):
        try:
            untracked_file_list = repo_.untracked_files
        except (UnicodeDecodeError, UnicodeEncodeError):
            # XXX: error in git python
            non_unicode_files = []
            non_unicode_dirs = []

            def has_unicode_error(name):
                try:
                    name.encode('ascii').decode('unicode_escape')\
                                                            .encode('latin1')
                    return False
                except (UnicodeDecodeError, UnicodeEncodeError):
                    return True
            #~ def has_unicode_error()

            for root_, dirs_, files_ in os.walk(repo_.working_tree_dir):
                for sub_d in dirs_:
                    sd_path = os.path.join(root_, sub_d)
                    if has_unicode_error(sd_path):
                        non_unicode_dirs.append(sd_path)
                for f_ in files_:
                    f_path = os.path.join(root_, f_)
                    if has_unicode_error(f_path):
                        non_unicode_files.append(f_path)
            if common_mix.confirm_execution("Do you want to delete non "
                                            "unicode files and dirs?"):
                # XXX: delete non unicode files
                for fd in non_unicode_dirs + non_unicode_files:
                    if os.path.isdir(fd):
                        func = shutil.rmtree
                    elif os.path.isfile(fd):
                        func = os.remove
                    elif os.path.islink(fd):
                        func = os.unlink
                    else:
                        ERROR_HANDLER.error_exit(\
                                "fd is not file, dir or link: {}".format(fd))

                    try:
                        func(fd)
                    except PermissionError:
                        # TODO: avoid using os.system here
                        os.system('sudo chmod -R 777 {}'.format(\
                                                        os.path.dirname(fd)))
                        func(fd)
                logging.debug("The following files and dirs were removed: {}"\
                                .format(non_unicode_files + non_unicode_dirs))
            else:
                ERROR_HANDLER.error_exit("Non unicode file name of untracked "
                            "files in the repo. Fix it and rerun", __file__)
            untracked_file_list = repo_.untracked_files
        res = set(untracked_file_list) | set(self._get_diffed(repo_))
        return res
    #~ def _get_untracked_and_diffed()

    def _get_diffed (self, repo_):
        res = {}
        d_ind = repo_.index.diff(None)
        for change_type in d_ind.change_type:
            for d in d_ind.iter_change_type(change_type):
                res[d.b_path] = d.change_type
        return res
    #~ def _get_diffed ()

    def _make_testing_branch(self, repo_dir, branch_name):
        # create 'test' branch if it doesn't exist 
        # so that it can be used for tests in this module
        clean_branch_list = self._get_branches_list(repo_dir)

        repo = git_repo(repo_dir)
        gitobj = repo.git
        
        if branch_name in clean_branch_list:
            newly_made = False
        else:
            gitobj.branch(branch_name)
            newly_made = True
        return newly_made
    #~ def _make_testing_branch()

    def _delete_testing_branch(self, repo_dir, branch_name):
        clean_branch_list = self._get_branches_list(repo_dir)

        repo = git_repo(repo_dir)
        gitobj = repo.git
        
        just_deleted = False
        if branch_name in clean_branch_list:
            gitobj.branch('-d', branch_name)
            just_deleted = True
        return just_deleted
    #~ def _delete_testing_branch()

    def _get_branches_list(self, repo_dir):
        repo = git_repo(repo_dir)
        gitobj = repo.git

        git_branch_string = gitobj.branch()
        git_branch_list = git_branch_string.split("\n")
        clean_branch_list = []
        for branch in git_branch_list:
            branch = branch.replace('*', '')
            branch = branch.replace(' ', '')
            clean_branch_list.append(branch)
        return clean_branch_list
    #~ def _get_branches_list()
#~ class RepositoryManager



""" This module implements the definition of the different configuration.
    NOTE: Make sure to run the function 'save_common_default_template'
        to save the default config after altering CompleteConfiguration.
        Write the default raw configuration template
            >>> import muteria.configmanager.configurations as mcc
            >>> mcc.save_common_default_template()

    We have a class defined for each configuration type. namely:
        - `ExecutionConfig`: Execution Configuration.
        - `ReportingConfig`: Reporting Configuration.
        - `ProjectConfig`: Project Configuration.
        - `TestcaseToolsConfig`: Testcase Tools Configuration.
        - `CodecoverageToolsConfig`: Codecoverage Tools Configuration.
        - `MutationToolsConfig`: Mutation Tools Configuration.

    TODO: Implement the restriction in adding extra element after creation
    TODO: Implement the verification of the parameters of each config
    TODO: Implement loading the values of the parameters from files... and 
            Checking
"""
#TODO: Implement test tool ordering (who use whose tests)

from __future__ import print_function

import os
import logging

import muteria.common.mix as common_mix 

from muteria.drivers.testgeneration import TestToolType

ERROR_HANDLER = common_mix.ErrorHandler

# TODO Find way to make the classes so that new elements cannot be added on the fly

class SessionMode(common_mix.EnumAutoName):
    EXECUTE_MODE = 0
    VIEW_MODE = 1
    INTERNAL_MODE = 2
    RESTORE_REPOS_MODE = 3
    CUSTOM_EXECUTION_MODE = 4
#~ class SessionMode

class ConfigClasses(common_mix.EnumAutoName):
    CONTROLLER_CONF = "CONTROLLER_CONF"
    PROJECT_CONF = "PROJECT_CONF"
    TESTCASES_CONF = "TESTCASES_CONF"
    CRITERIA_CONF = "CRITERIA_CONF"
    OTHER_CONF = "OTHER_CONF"
#~ class ConfigClasses

class ConfigElement(object):
    def __init__(self, val=None, desc=None, val_range=None, conf_class=None):
        self.val = val
        self.desc = desc
        self.val_range = val_range
        self.conf_class = conf_class
    #~ def __self__()

    def set_val(self, new_val):
        self.val = new_val
    def get_val(self):
        return self.val
    
    def set_desc(self, new_desc):
        self.desc = new_desc
    def get_desc(self):
        return self.desc

    def set_val_range(self, new_val_range):
        self.val_range = new_val_range
    def get_val_range(self):
        return self.val_range

    def set_conf_class(self, new_conf_class):
        self.conf_class = new_conf_class
    def get_conf_class(self):
        return self.conf_class
#~ class ConfigElement()

class CompleteConfiguration(object):

    #######################################################
    #######             Execution Parameters         ######
    #######################################################
    # Criteria to Include in the analysis
    # The corresponding configs of codecoverage (`CodecoverageToolsConfig`)
    # and mutation (`MutationToolsConfig`) are considered if and only if 
    # the corresponding include is enabled

    # Decide whether to start over, deleting previous, for execution mode
    EXECUTION_CLEANSTART=False

    # Value of type SessionMode (Mandatory)
    RUN_MODE = None

    # List of enabled criteria 
    # (must have tool specified is specifically enabled)
    # None mean all criteria with tool specified
    ENABLED_CRITERIA = [] 

    # Enable keeping output summary for program passfail execution
    GET_PASSFAIL_OUTPUT_SUMMARY = True

    # keepoutput summary for the following criteria (may run slower)
    CRITERIA_WITH_OUTPUT_SUMMARY = []

    # Decides whether to hash the output log
    HASH_OUTLOG = True

    # PARALELISM
    SINGLE_REPO_PARALLELISM = 1 # Max number of parallel exec in a repo dir

    # MICRO CONTROLS
    EXECUTE_ONLY_CURENT_CHECKPOINT_META_TASK = False # for Debugging
    RESTART_CURRENT_EXECUTING_META_TASKS = False
    # Specify a Step to go back to
    RE_EXECUTE_FROM_CHECKPOINT_META_TASKS = [] # Make interaction easy

    # Output dir pathname (Mandatory)
    OUTPUT_ROOT_DIR = None

    # Enable logging debug data
    LOG_DEBUG = False

    #######################################################
    #######             Reporting Parameters         ######
    #######################################################
    GENERATE_HTML_REPORT = True

    OUTPUT_CRITERIA_SCORES = True
    CRITERIA_SCORE_BY_TOOL = True
    OUTPUT_CRITERIA_SUBSUMPTION_SCORE = True

    OUTPUT_CRITERIA_COVERED_ELEMENTS = False
    OUTPUT_CRITERIA_UNCOVERED_ELEMENTS = True
    DETAILED_ELEMENTS_OUTPUT = False

    OUTPUT_CRITERIA_SUBSUMING_ELEM_NUM = True

    # When iteratively try to cover some element, show 
    OUTPUT_STATS_HISTORY = True

    #######################################################
    #######        Project Config Parameters         ######
    #######################################################
    # Project programming language (Mandatory)
    PROGRAMMING_LANGUAGE = None

    # Repository dir pathname (Mandatory)
    REPOSITORY_ROOT_DIR = None
    
    # string representing the relative path to the executable
    # (or entry point file) in the repository
    REPO_EXECUTABLE_RELATIVE_PATHS = None

    # optional map between each source and the corresponding intermediate
    # file map (such as object or assembly file for c of .class file for java)
    TARGET_SOURCE_INTERMEDIATE_CODE_MAP = None

    # Specify the optional general scope of the evaluation, 
    # Specific scope can be specified per tool
    TARGET_CLASSES_NAMES = None
    # None value mean all functions
    TARGET_METHODS_BY_TARGET_CLASSES = None
    # each source file element is the relative path from repos rootdir.
    # None value means all source files
    ##TARGET_SOURCE_FILES_LIST = None

    # each test file element is the relative path from repos rootdir.
    # None value means all dev tests
    DEVELOPER_TESTS_LIST = None

    # Function that takes 3 arguments:
    #   <test_name: str>
    #   <repos directory rootdir: str>
    #   <Executable relative path map: dict>
    #   <env_vars: map>
    #   <timeout: int>
    # and run with the executable as in repository
    # The function return:
    #   0 on passing test
    #   1 on failing test
    #   -1 on error
    CUSTOM_DEV_TEST_RUNNER_FUNCTION = None

    CUSTOM_DEV_TEST_PROGRAM_WRAPPER_CLASS = None

    # Take the outlog as a string and return the cleaned version
    CUSTOM_TEST_EXECUTION_OUTPUT_CLEANER_FUNCTION = None

    # Optional. When not None, the CUSTOM_DEV_TEST_RUNNER is the name of 
    # the function in this file to use
    CUSTOM_DEV_TEST_RUNNER_MODULE = None

    # Function that build the code to execute 
    # (for compiled languages such as C)
    # The function has 5 parameters: 
    #   <repos directory rootdir: str>
    #   <Executable relative path: str>
    #   <Optional compiler to use(compiler name): str> 
    #   <optional flags to pass to compiler: list of flags> 
    #   <clean temporary before build: bool>
    #   <reconfigure before build: bool>
    # And returns:
    #   True on success  
    #   False on failure
    # A code builder class is defined and make use of this function 
    # It ensure on call to this functin at the time and no call while any 
    # processing in the repodir. 
    CODE_BUILDER_FUNCTION = None

    # Optional. When not None, the CODE_BUILDER_FUNC is the name of 
    # the function in this file to use
    CODE_BUILDER_MODULE = None

    #######################################################
    #######             Tools parameters             ######
    #######################################################
    # ===================== TESTCASES =====================#
    # Tests already existing before execution starts
    DEVELOPER_TESTS_ENABLED = True
    # Test created during execution
    GENERATED_TESTS_ENABLED = True

    STOP_TESTS_EXECUTION_ON_FAILURE = False # Will not get full matrix

    DISCARD_FLAKY_TESTS = True

    # Test tool types in order of execution (Test generation). 
    # elements in same tuple have parallel execution. elements at tuple i
    # generate tests after element at i-1 and may use results of test from i-1.
    TEST_TOOL_TYPES_SCHEDULING = [
        (TestToolType.USE_ONLY_CODE,), 
        (TestToolType.USE_CODE_AND_TESTS,),
    ]

    # Test generation tool. Example:
    # >>> TestcaseToolsConfig(tooltype=TestToolType.USE_ONLY_CODE,
    #                        toolname='klee', config_id=0)
    TESTCASE_TOOLS_CONFIGS = [

    ]

    TESTCASES_SELECTION = None

    # Reporting
    REPORT_NUMBER_OF_TESTS_GENERATED = True
    REPORT_NUMBER_OF_DUPLICATED_TESTS = True

    ## --- Modifiable (Testcase) ---##
    # use test case oracle as oracle
    #TESTS_ORACLE_TESTS = True
    # Use output of the specified version as oracle, 
    # Pass filepath to repo patch 
    #TESTS_ORACLE_OTHER_VERSION = None
    # file path to an executable to use as oracle
    #TESTS_ORACLE_OTHER_EXECUTABLE = None

    #TEST_GENERATION_TIMEOUT = 7200.0 # in seconds
    #ONE_TEST_EXECUTION_TIMEOUT = 900.0 # in seconds (Handle inifnite loops)
    # ========================================================#

    # ===================== CRITERIA COVERAGE =====================#

    # Map with key criteria and values the list of tools
    # criteria tool. Example:
    # >>> CriteriaToolConfig(tooltype=CriteriaToolType.USE_ONLY_CODE,
    #                        toolname='gcov', config_id=0)
    CRITERIA_TOOLS_CONFIGS_BY_CRITERIA = {

    }

    # List of sets of test criteria stating the order in which criteria 
    # should be executed. Example strom mutation after weak mutation.
    # When None, Default to the order in:
    # >>> muteria.drivers.criteria.CRITERIA_SEQUENCE
    CRITERIA_SEQUENCE = None

    # List of criteria that have test objectives covered when test execution
    # differs with original
    # When None, Default to the order in:
    # >>> muteria.drivers.criteria.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM
    CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM = None

    # List of criteria to run with failing tests
    RUN_FAILING_TESTS_WITH_CRITERIA = [

    ]

    # List of criteria to run with passing tests
    RUN_PASSING_TESTS_WITH_CRITERIA = [

    ]

    CRITERIA_RESTRICTION_ENABLED = True  # Enable restricting mutation(scope)

    # criterion: selection tools. Example: SM and TCE or E-SELECTIVE
    CRITERIA_ELEM_SELECTIONS = {
        
    }

    ONLY_EXECUTE_SELECTED_CRITERIA_ELEM = True

    MAX_CRITERIA_ELEM_SELECTION_NUM_PERCENT = '100%'

    # Criterion: guider dict. ex: {STRONG_MUTATION: Surviving}
    CRITERIA_TESTGEN_GUIDANCE = {

    }

    # Criterion: optimizers dict. ex: {STRONG_MUTATION: SM_OPTIMIZED_BY_WM}
    CRITERIA_EXECUTION_OPTIMIZERS = {

    }

    # Will not get full matrix. the non executed will be uncertain
    COVER_CRITERIA_ELEMENTS_ONCE = False 

    ## --- Modifiable (Code) ---##
    #SEPARATED_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES = 60.0 # in seconds
    #META_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES = 600.0 # in seconds
    # ========================================================#

    #######################################################
    #######             Extra parameters             ######
    #######################################################
    #LLVM_TO_NATIVE_LINKING_FLAGS = None
#~ class CompleteConfiguration

class BaseToolConfig(dict):
    """
    :param criteria_on: (list) alternate way to represent criteria of a tool (TODO)
    """
    def __init__(self, tooltype=None, toolname=None, config_id=None, \
                                    criteria_on=None, tool_user_custom=None):
        self.tooltype = tooltype
        self.toolname = toolname
        self.config_id = config_id
        self.criteria_on = criteria_on
        self.tool_user_custom=tool_user_custom
        if self.config_id is None:
            self.toolalias = self.toolname
        else:
            self.toolalias = toolname+"_"+str(self.config_id)
    #~ def __init__()

    def __eq__(self, value):
        return (self.__dict__ == value.__dict__)

    def get_tool_name(self):
        return self.toolname
    def get_tool_config_alias(self):
        return self.toolalias
    def get_tool_type(self):
        return self.tooltype
    def get_tool_criteria_on(self):
        return self.criteria_on
    def get_tool_user_custom(self):
        return self.tool_user_custom
#~ class BaseToolConfig

class ToolUserCustom(dict):
    '''
    This config file is helpful to specify tool specific configurations
    For example:
    >>> conf = ToolUserCustom( \
        PATH_TO_TOOL_EXECUTABLE='/fullpath/to/tool/dir', \
        ENV_VARS_DICT={'var1': 'value1'}, \
        PRE_TARGET_CMD_ORDERED_FLAGS_LIST=[('-solver', 'stp'), \
                    ('-mutation-scope', os.path.abspath("scopefile"))], \
        POST_TARGET_CMD_ORDERED_FLAGS_LIST=[('-sym-args', '0', '2', '3')]
    )

    Note: It is possibled to only specify a subset of the values.
    '''
    # -str- custom path to executable parent dir to use for tool.
    # Directory containing executables (fullpath: string)
    # Useful especially for tools having many executables like 'klee'
    PATH_TO_TOOL_BINARY_DIR = None
    # -dict- key value of <environment var: var value> both strings
    ENV_VARS_DICT = None
    # -list of tuples- ordered list of tuples, each tuple represent a flag
    # as first element and its values. ordering is kept
    # example: for GNU find, we can have:
    # [('-type', 'f'), ('-exec', 'grep', '{}', '\\;')]
    # PRE_TARGET_... is the arguments list that must go before the 
    # analyzed program (target) in command line order POST_TARGET go after. 
    PRE_TARGET_CMD_ORDERED_FLAGS_LIST = None
    POST_TARGET_CMD_ORDERED_FLAGS_LIST = None

    def __init__(self, **kwargs):
        for k in kwargs:
            ERROR_HANDLER.assert_true(k.isupper() and k in dir(self), \
                                    "invalid parameter passed: "+k, __file__)
            setattr(self, k, kwargs[k])
        #num_elem = len(self)
        #super(ToolUserCustom, self).__init__(**kwargs)
        #self.__dict__ = self
        #diff = len(self.__dict__) - num_elem
        #if diff != 0:
        #    ERROR_HANDLER.error_exit(\
        #                "{} invalid parameter was/were added".format(diff), \
        #                                                            __file__)
    #~def __init__()

    def __eq__(self, value):
        return (self.__dict__ == value.__dict__)
#~class ToolUserCustom


class TestcaseToolsConfig(BaseToolConfig):
    # use test case oracle as oracle
    TESTS_ORACLE_TESTS = True
    # Use output of the specified version as oracle, 
    # Pass filepath to repo patch 
    TESTS_ORACLE_OTHER_VERSION = None
    # file path to an executable to use as oracle
    TESTS_ORACLE_OTHER_EXECUTABLE = None

    TEST_GENERATION_MAXTIME = 7200.0 # in seconds
    ONE_TEST_EXECUTION_TIMEOUT = 60.0 # in seconds (Handle inifnite loops)

    # consider test execution error as a test failure
    TEST_EXECUTION_ERROR_AS_FAIL = False
    
    def set_test_gen_maxtime(self, max_time):
        self.TEST_GENERATION_MAXTIME = max_time
    def set_one_test_execution_timeout(self, timeout):
        self.ONE_TEST_EXECUTION_TIMEOUT = timeout
    def set_test_oracle_test(self, value):
        self.TESTS_ORACLE_TESTS = value
    def set_test_execution_error_as_failure(self, value):
        self.TEST_EXECUTION_ERROR_AS_FAIL = value
#~class TestcaseToolsConfig
    
class CriteriaToolsConfig(BaseToolConfig):
    SEPARATED_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES = 1.5 # proportion
    META_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES = 100.0 # proportion
#~class CriteriaToolsConfig



#-------------------------------------------------#
# DEFAULT CONF TEMPLATE CREATION HELPERS          #
# Use these to generate/update default raw config #
#-------------------------------------------------#

def get_full_rawconf_template():
    """ Computes the defaul raw configuration template as string list
    """
    thisfile = os.path.abspath(__file__)
    with open(thisfile) as f:
        # read all content of the class CompleteConfiguration into list
        row_list = []
        active = False
        for line in f:
            stripped_line = line.strip()
            if not active:
                if stripped_line == "class CompleteConfiguration(object):":
                    active = True
            else:
                if stripped_line == "#~ class CompleteConfiguration":
                    active = False
                else:
                    row_list.append(stripped_line)

    # Check that it was fine
    if len(row_list) < 1:
        ERROR_HANDLER.error_exit(\
                    'complete configuration class was not found', __file__)
    ERROR_HANDLER.assert_true(not active, \
                    "did not find end of configuration class", __file__)
    return row_list
#~ def get_full_rawconf_template()

def save_common_default_template(filename=None):
    """ Write the default raw configuration template
        >>> import muteria.configmanager.configurations as mcc
        >>> mcc.save_common_default_template()
    """
    if filename is None:
        filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),\
                                            "defaults", "common_defaults.py")
    row_list = get_full_rawconf_template()
    # save into out file
    if os.path.isfile(filename):
        if not common_mix.confirm_execution(\
                        "The filename to save configuration template "
                        "exists. Do you want to override it?"):
            return 
    

    header = "\"\"\" Defaults parameters that are common to all languages\n" +\
            "\"\"\"\n"
    header += "from __future__ import print_function\n\n"
    header += "{} {}\n".format("from muteria.configmanager.configurations", \
                                "import SessionMode")

    header += "{} {}\n".format("from muteria.configmanager.configurations", \
                                "import TestcaseToolsConfig")
    header += "{} {}\n".format("from muteria.configmanager.configurations", \
                                "import CriteriaToolsConfig")
    header += "{} {}\n".format("from muteria.configmanager.configurations", \
                                "import ToolUserCustom")

    header += "\n{} {}\n".format("from muteria.drivers.criteria", \
                                "import TestCriteria")
    header += "{} {}\n".format("from muteria.drivers.criteria", \
                                "import CriteriaToolType")
    header += "\n{} {}\n".format("from muteria.drivers.testgeneration", \
                                "import TestToolType")
    with open(filename, 'w') as f:
        f.write(header+'\n')
        for row in row_list:
            f.write(row+'\n')
#~ def save_common_default_template()

if __name__ == '__main__':
    save_common_default_template()


#
# [LICENCE]
#
""" Testcase tool module. The class of interest is BaseTestcaseTool.
The class represent a testcase tool which manages a working directory
where, given code files(executable) represented in the repository dir,
corresponding tests are generated (harvested) by code files. Those test can
also be executed using the same code used to generate the tests or 
other code specified. 

The tools are organized by programming language

For each language, there is a folder for each tool, 
named after the tool in all lowercase , starting with letter or underscore(_),
The remaining caracters are either letter, number or underscore

XXX Each testcase tool package must make each test case tool visible in 
the __init__.py file (Just set to None if not defined).
For Example: 
- If nothing is defined:
>>> StaticTestcaseTool = None
>>> StaticMutantTestcaseTool = None
>>> DynamicTestcaseTool = None
>>> DynamicMutantTestcaseTool = None

- If  for example the StaticTestcaseTool is defined as the class 
<class extending BaseTestcaseTool>, in th module <Module>, replace
 "StaticTestcaseTool = None" with:
>>> import <Module>.<class extending BaseTestcaseTool> as StaticTestcaseTool

"""

from __future__ import print_function
import os
import sys
import glob
import shutil
import logging
import abc
import hashlib
import time
import multiprocessing
import joblib

import tqdm

import muteria.common.matrices as common_matrices
import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.drivers import DriversUtils

from muteria.drivers.checkpoint_handler import CheckPointHandler
from muteria.repositoryandcode.callback_object import DefaultCallbackObject

ERROR_HANDLER = common_mix.ErrorHandler

class BaseTestcaseTool(abc.ABC):
    '''
    '''

    def __init__(self, tests_working_dir, code_builds_factory, config, \
                                        head_explorer, checkpointer, \
                                        parent_meta_tool=None):
        # Set Constants
        self.compress_test_storage_dir = True

        # Set Direct Arguments Variables
        self.tests_working_dir = tests_working_dir
        self.code_builds_factory = code_builds_factory
        self.config = config
        self.head_explorer = head_explorer
        self.checkpointer = checkpointer
        self.parent_meta_tool = parent_meta_tool

        # Verify Direct Arguments Variables
        ERROR_HANDLER.assert_true(self.tests_working_dir is not None, \
                                    "Must specify tests_working_dir", __file__)

        # Set Indirect Arguments Variables
        ## Generate the tests into this folder (to be created by user)
        self.tests_storage_dir = os.path.join(
                        self.tests_working_dir, "tests_files")
        self.tests_storage_dir_archive = os.path.join(
                        self.tests_working_dir, "tests_files.tar.gz")
        self.custom_binary_dir = None
        if self.config.tool_user_custom is not None:
            self.custom_binary_dir = \
                        self.config.tool_user_custom.PATH_TO_TOOL_BINARY_DIR
            ERROR_HANDLER.assert_true(self.custom_binary_dir is None or \
                                    os.path.isdir(self.custom_binary_dir), \
                            "specified custom_binary_dir does not exist", \
                                                                    __file__)

        # Verify indirect Arguments Variables

        # Initialize Other Fields
        self.test_execution_time = {}
        self.test_execution_time_storage_file = os.path.join(\
                        self.tests_working_dir, "test_to_execution_time.json")
        self.shared_loc = multiprocessing.RLock()

        # Make Initialization Computation
        ## Create dirs
        if not os.path.isdir(self.tests_working_dir):
            self.clear_working_dir()

        if os.path.isfile(self.test_execution_time_storage_file):
            self.test_execution_time = common_fs.loadJSON(\
                                        self.test_execution_time_storage_file)

        # decompress potential test storage archive
        if self.compress_test_storage_dir:
            if os.path.isfile(self.tests_storage_dir_archive):
                if os.path.isdir(self.tests_storage_dir):
                    try:
                        shutil.rmtree(self.tests_storage_dir)
                    except PermissionError:
                        self._dir_chmod777(self.tests_storage_dir)
                        shutil.rmtree(self.tests_storage_dir)
                common_fs.TarGz.decompressDir(self.tests_storage_dir_archive)
    #~ def __init__()

    def __del__(self):
        # remove the decompressed the test storage dir
        if self.compress_test_storage_dir:
            if os.path.isfile(self.tests_storage_dir_archive) \
                                    and os.path.isdir(self.tests_storage_dir):
                shutil.rmtree(self.tests_storage_dir)
    #~ def __del__()

    def clear_working_dir(self):
        if os.path.isdir(self.tests_working_dir):
            shutil.rmtree(self.tests_working_dir)
        os.mkdir(self.tests_working_dir)
    #~ def clear_working_dir(self):

    def get_toolname(self):
        return self.config.get_tool_name()
    #~ def get_toolname()

    def get_toolalias(self):
        return self.config.get_tool_config_alias()
    #~ def get_toolalias()

    def get_checkpointer(self):
        return self.checkpointer
    #~ def get_checkpointer()

    def has_checkpointer(self):
        return self.checkpointer is not None
    #~ def has_checkpointer(self)

    def execute_testcase (self, testcase, exe_path_map, env_vars, \
                                        timeout=None, \
                                        use_recorded_timeout_times=None, \
                                        recalculate_execution_times=False, \
                                        with_output_summary=True, \
                                        hash_outlog=True):
        return self._execute_testcase(testcase, exe_path_map, env_vars, \
                                timeout=timeout, \
                                use_recorded_timeout_times=\
                                                use_recorded_timeout_times, \
                                recalculate_execution_times=\
                                                recalculate_execution_times, \
                                with_output_summary=with_output_summary, \
                                hash_outlog=hash_outlog)
    #~ def execute_testcase()

    def runtests(self, testcases, exe_path_map, env_vars, \
                                stop_on_failure=False, per_test_timeout=None, \
                                use_recorded_timeout_times=None, \
                                recalculate_execution_times=False, \
                                with_output_summary=True, hash_outlog=True, \
                                parallel_count=1):
        return self._runtests(testcases=testcases, exe_path_map=exe_path_map, \
                                env_vars=env_vars, \
                                stop_on_failure=stop_on_failure, \
                                per_test_timeout=per_test_timeout, \
                                use_recorded_timeout_times=\
                                                use_recorded_timeout_times, \
                                recalculate_execution_times=\
                                                recalculate_execution_times, \
                                with_output_summary=with_output_summary, \
                                hash_outlog=hash_outlog, \
                                parallel_count=parallel_count)
    #~ def runtests()
                            
    class RepoRuntestsCallbackObject(DefaultCallbackObject):
        def after_command(self):
            # Copy the exes into the repo (user must revert)
            exe_path_map = self.post_callback_args[1]['exe_path_map']
            if self.post_callback_args[2]:
                self._copy_to_repo(exe_path_map, skip_none_dest=True)

            # execute
            res = self.post_callback_args[0](**self.post_callback_args[1])
            return res
    #~ class RepoRuntestsCallbackObject

    def _in_repo_execute_testcase(self, testcase, exe_path_map, env_vars, \
                                        timeout=None, \
                                        use_recorded_timeout_times=None, \
                                        recalculate_execution_times=False, \
                                        with_output_summary=True, \
                                        hash_outlog=True, \
                                        copy_exe_to_repo=True):
        callback_func = self._execute_testcase
        cb_obj = self.RepoRuntestsCallbackObject()
        cb_obj.set_post_callback_args((callback_func,
                                {
                                    "testcase": testcase,
                                    "exe_path_map": exe_path_map,
                                    "env_vars": env_vars,
                                    "timeout": timeout,
                                    "use_recorded_timeout_times":\
                                            use_recorded_timeout_times, \
                                    "recalculate_execution_times":\
                                            recalculate_execution_times, \
                                    "with_output_summary":with_output_summary,\
                                    "hash_outlog":hash_outlog,
                                }, copy_exe_to_repo))
        repo_mgr = self.code_builds_factory.repository_manager
        _, exec_verdict = repo_mgr.custom_read_access(cb_obj)
        # revert exes
        self.code_builds_factory.set_repo_to_build_default()
        return exec_verdict
    #~ def _in_repo_execute_testcase()

    def _in_repo_runtests(self, testcases, exe_path_map, env_vars, \
                                stop_on_failure=False, per_test_timeout=None, \
                                use_recorded_timeout_times=None, \
                                recalculate_execution_times=False, \
                                with_output_summary=True, hash_outlog=True, \
                                parallel_count=1, \
                                copy_exe_to_repo=True):
        callback_func = self._runtests
        cb_obj = self.RepoRuntestsCallbackObject()
        cb_obj.set_post_callback_args((callback_func,
                                {
                                    "testcases": testcases,
                                    "exe_path_map": exe_path_map,
                                    "env_vars": env_vars,
                                    "stop_on_failure": stop_on_failure,
                                    "per_test_timeout": per_test_timeout,
                                    "use_recorded_timeout_times":\
                                            use_recorded_timeout_times, \
                                    "recalculate_execution_times":\
                                            recalculate_execution_times, \
                                    "with_output_summary":with_output_summary,\
                                    "hash_outlog":hash_outlog, \
                                    "parallel_count": parallel_count,
                                }, copy_exe_to_repo))
        repo_mgr = self.code_builds_factory.repository_manager
        _, exec_verdicts = repo_mgr.custom_read_access(cb_obj)
        # revert exes
        self.code_builds_factory.set_repo_to_build_default()
        return exec_verdicts
    #~ def _in_repo_runtests()

    def _execute_testcase (self, testcase, exe_path_map, env_vars, \
                                        timeout=None, \
                                        use_recorded_timeout_times=None, \
                                        recalculate_execution_times=False, \
                                        with_output_summary=True, \
                                        hash_outlog=True):
        '''
        Execute a test case with the given executable and 
        say whether it failed

        :param testcase: string name of the test cases to execute
        :param exe_path_map: string representing the file system path to 
                        the executable to execute with the test
        :param env_vars: dict of environment variables to set before
                        executing the test ({<variable>: <value>})
        :returns: boolean failed verdict of the test 
                        (True if failed, False otherwise)
        '''

        if timeout is None:
            if use_recorded_timeout_times is not None:
                ERROR_HANDLER.assert_true(use_recorded_timeout_times > 0, \
                                        "use_recorded_timeout_times must be "
                                        "positive if not None", __file__)
                if testcase in self.test_execution_time:
                    timeout = self.test_execution_time[testcase] * \
                                                use_recorded_timeout_times
        else:
            ERROR_HANDLER.assert_true(use_recorded_timeout_times is None, \
                                "use_recorded_timeout_times must not be set "
                                "when timeout is not None", __file__)


        self._prepare_executable(exe_path_map, env_vars, \
                                            collect_output=with_output_summary)
        self._set_env_vars(env_vars)

        start_time = time.time()
        fail_verdict, execoutlog_hash = \
                        self._oracle_execute_a_test(testcase, exe_path_map, \
                                            env_vars, timeout=timeout, \
                                    with_output_summary=with_output_summary, \
                                                    hash_outlog=hash_outlog)

        # Record exec time if not existing
        if recalculate_execution_times:
            self.test_execution_time[testcase] = \
                                        1 + int(time.time() - start_time)

        self._restore_env_vars()
        self._restore_default_executable(exe_path_map, env_vars, \
                                            collect_output=with_output_summary)

        return fail_verdict, execoutlog_hash
    #~ def _execute_testcase()

    def _runtests(self, testcases, exe_path_map, env_vars, \
                                stop_on_failure=False, per_test_timeout=None, \
                                use_recorded_timeout_times=None, \
                                recalculate_execution_times=False, \
                                with_output_summary=True, hash_outlog=True, \
                                parallel_count=1):
        '''
        Execute the list of test cases with the given executable and 
        say, for each test case, whether it failed.
        Note: Re-implement this if there the tool implements ways to faster
        execute multiple test cases.

        :param testcases: list of test cases to execute
        :param exe_path_map: string representing the file system path to 
                        the executable to execute with the tests
        :param env_vars: dict of environment variables to set before
                        executing each test ({<variable>: <value>})
        :param stop_on_failure: decide whether to stop the test execution once
                        a test fails
        :returns: plitair of:
                - dict of testcase and their failed verdict.
                 {<test case name>: <True if failed, False if passed, 
                    UNCERTAIN_TEST_VERDICT if uncertain>}
                 If stop_on_failure is True, only return the tests that have 
                 been executed until the failure
                - test execution output log hash data object or None
        '''
        # @Checkpoint: create a checkpoint handler (for time)
        checkpoint_handler = CheckPointHandler(self.get_checkpointer())
        if checkpoint_handler.is_finished():
            logging.warning("{} {} {}".format( \
                            "The function 'runtests' is finished according", \
                            "to checkpoint, but called again. None returned", \
                            "\nPlease Confirm reexecution..."))
            if common_mix.confirm_execution("{} {}".format( \
                                "Function 'runtests' is already", \
                                "finished, do yo want to restart?")):
                checkpoint_handler.restart()
                logging.info("Restarting the finished 'runtests'")
            else:
                ERROR_HANDLER.error_exit(err_string="{} {} {}".format( \
                        "Execution halted. Cannot continue because no value", \
                        " can be returned. Check the results of the", \
                        "finished execution"), call_location=__file__)

        if per_test_timeout is None:
            per_test_timeout = {tc: None for tc in testcases}
            if use_recorded_timeout_times is not None:
                ERROR_HANDLER.assert_true(use_recorded_timeout_times > 0, \
                                        "use_recorded_timeout_times must be "
                                        "positive if not None", __file__)
                per_test_timeout.update({x: (y * use_recorded_timeout_times) \
                                for x, y in self.test_execution_time.items()})
        else:
            ERROR_HANDLER.assert_true(use_recorded_timeout_times is None, \
                                "use_recorded_timeout_times must not be set "
                                "when per_test_timeout is set", __file__)

        # Prepare the exes
        self._prepare_executable(exe_path_map, env_vars, \
                                            collect_output=with_output_summary)
        self._set_env_vars(env_vars)

        test_failed_verdicts = {} 
        test_outlog_hash = {} 
        processbar = tqdm.tqdm(testcases, leave=False, dynamic_ncols=True) 

        # Parallel stuffs
        ptest_tresh = 5

        def test_exec_iteration(testcase):
            processbar.set_description("Running Test %s"% testcase)
            start_time = time.time()
            test_failed, execoutlog_hash = \
                        self._oracle_execute_a_test(testcase, exe_path_map, \
                                        env_vars, \
                                        timeout=per_test_timeout[testcase], \
                                    with_output_summary=with_output_summary, \
                                        hash_outlog=hash_outlog)
            
            # Record exec time if not existing
            with self.shared_loc:
                if recalculate_execution_times:
                    self.test_execution_time[testcase] = \
                                            1 + int(time.time() - start_time)

                test_failed_verdicts[testcase] = test_failed
                test_outlog_hash[testcase] = execoutlog_hash
            return test_failed
        #~ def test_exec_iteration()

        if self.can_run_tests_in_parallel() and len(testcases) >= ptest_tresh \
                        and parallel_count is not None and parallel_count > 1:
            paralle_count = min(len(testcases), parallel_count)
            joblib.Parallel(n_jobs=paralle_count, require='sharedmem')\
                            (joblib.delayed(test_exec_iteration)(testcase) \
                                                for testcase in processbar)
        else:
            for testcase in processbar: 
                test_failed = test_exec_iteration(testcase)
                if stop_on_failure and test_failed != \
                                common_mix.GlobalConstants.PASS_TEST_VERDICT:
                    break
        
        if recalculate_execution_times:
            common_fs.dumpJSON(self.test_execution_time, \
                            self.test_execution_time_storage_file, pretty=True)

        # Restore back the exes
        self._restore_env_vars()
        self._restore_default_executable(exe_path_map, env_vars, \
                                            collect_output=with_output_summary)

        if stop_on_failure:
            # Make sure the non executed test has the uncertain value (None)
            if len(test_failed_verdicts) < len(testcases):
                for testcase in set(testcases) - set(test_failed_verdicts):
                    test_failed_verdicts[testcase] = \
                            common_mix.GlobalConstants.UNCERTAIN_TEST_VERDICT
                    test_outlog_hash[testcase] = common_matrices.\
                                        OutputLogData.UNCERTAIN_TEST_OUTLOGDATA

        # @Checkpoint: Finished (for time)
        checkpoint_handler.set_finished(None)

        if not with_output_summary:
            test_outlog_hash = None

        return test_failed_verdicts, test_outlog_hash
    #~ def _runtests()

    def _oracle_execute_a_test (self, testcase, exe_path_map, env_vars, \
                                        callback_object=None, timeout=None,
                                with_output_summary=True, hash_outlog=True):
        """ Execute a test and use the specified oracles to check
            Also collect the output

            :param hash_outlog: (bool) Choose to hash or not at runtime 
                                (flakiness check)
        """

        if timeout is None:
            timeout = self.config.ONE_TEST_EXECUTION_TIMEOUT
        
        #logging.debug(str(timeout))
        verdict, output_err = self._execute_a_test(\
                                            testcase,exe_path_map, env_vars,\
                                            callback_object=callback_object, \
                                            timeout=timeout, \
                                            collect_output=with_output_summary)
        if with_output_summary:
            retcode, outlog, timedout = output_err
            outlog = self.code_builds_factory\
                                .repository_manager\
                                .get_test_exec_output_cleaner_func()(outlog)
            out_len = len(outlog)
            if hash_outlog:
                outlog = outlog.encode('utf-8', 'backslashreplace')
                out_hash_val = hashlib.sha512(outlog).hexdigest()
            else:
                out_hash_val = outlog
            outlog_summary = {
                common_matrices.OutputLogData.OUTLOG_LEN: out_len,
                common_matrices.OutputLogData.OUTLOG_HASH: out_hash_val,
                common_matrices.OutputLogData.RETURN_CODE: retcode,
                common_matrices.OutputLogData.TIMEDOUT: timedout,
            }
        else:
            outlog_summary = None

        return verdict, outlog_summary
    #~ def _oracle_execute_a_test()

    def generate_tests (self, exe_path_map, \
                            meta_criteria_tool_obj=None, \
                            parallel_count=1, \
                            code_builds_factory_override=None, max_time=None):
        '''
        '''

        logging.debug("# Generating tests with {} ...".format(\
                                        self.config.get_tool_config_alias()))

        # @Checkpoint: create a checkpoint handler (for time)
        checkpoint_handler = CheckPointHandler(self.get_checkpointer())
        if checkpoint_handler.is_finished():
            return

        outputdir = self.tests_storage_dir
        if code_builds_factory_override is None:
            code_builds_factory_override = self.code_builds_factory
        if os.path.isdir(outputdir):
            shutil.rmtree(outputdir)

        # If compressing test storage dir, remove archive if working on 
        # the default test storage dir
        if self.compress_test_storage_dir \
                                and not os.path.isdir(self.tests_storage_dir):
            if os.path.isfile(self.tests_storage_dir_archive):
                os.remove(self.tests_storage_dir_archive)

        os.mkdir(outputdir)
        self._do_generate_tests (exe_path_map, \
                            code_builds_factory=code_builds_factory_override, \
                            meta_criteria_tool_obj=meta_criteria_tool_obj, \
                                                            max_time=max_time)


        # Compress test storage dir?
        if self.compress_test_storage_dir:
            if os.path.isdir(self.tests_storage_dir):
                common_fs.TarGz.compressDir(self.tests_storage_dir, \
                                            self.tests_storage_dir_archive, \
                                            remove_in_directory=False)

        # @Checkpoint: Finished (for time)
        checkpoint_handler.set_finished(None)
    #~ def generate_tests()

    def _set_env_vars(self, env_vars):
        if env_vars:
            try:
                if self.env_vars_store is not None:
                    ERROR_HANDLER.error_exit(\
                            "Bug: env_var set again without restore", __file__)
            except AttributeError:
                pass
            self.env_vars_store = os.environ.copy()
            os.environ.update(env_vars)
        else:
            self.env_vars_store = os.environ
    #~ def _set_env_vars()

    def _restore_env_vars(self):
        try:
            if self.env_vars_store is None:
                raise AttributeError
        except AttributeError:
            ERROR_HANDLER.error_exit("restoring unset env")
        os.environ = self.env_vars_store
        self.env_vars_store = None
    #~ def _restore_env_vars()

    def tool_installed (self):
        """ Check that a tool with given conf is installed
        """
        return self.installed(custom_binary_dir=self.custom_binary_dir)
    #~ def tool_installed ()

    def _dir_chmod777(self, dirpath):
        try:
            for root_, dirs_, files_ in os.walk(dirpath):
                for sub_d in dirs_:
                    os.chmod(os.path.join(root_, sub_d), 0o777)
                for f_ in files_:
                    os.chmod(os.path.join(root_, f_), 0o777)
        except PermissionError:
            ret,out,_ = DriversUtils.execute_and_get_retcode_out_err('sudo', \
                                            ['chmod', '777', '-R', dirpath])
            ERROR_HANDLER.assert_true(ret == 0, \
                        "'sudo chmod 777 -R "+dirpath+"' failed (returned "+\
                                        str(ret)+"), error: "+out, __file__)
    #~ def _dir_chmod777()

    def requires_criteria_instrumented(self):
        return False
    #~ def requires_criteria_instrumented()

    def can_run_tests_in_parallel(self):
        return False
    #~ def can_run_tests_in_parallel()

    #######################################################################
    ##################### Methods to implement ############################
    #######################################################################

    @classmethod
    @abc.abstractclassmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def installed()

    @abc.abstractmethod
    def get_testcase_info_object(self):
        print ("!!! Must be implemented in child class !!!")
    #~ def get_testcase_info_object()

    @abc.abstractmethod
    def _prepare_executable (self, exe_path_map, env_vars, \
                                                        collect_output=False):
        """ Make sure we have the right executable ready (if needed)
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def _prepare_executale()

    @abc.abstractmethod
    def _restore_default_executable(self, exe_path_map, env_vars, \
                                                        collect_output=False):
        """ Restore back the default executable (if needed).
            Useful for test execution that require the executable
            at a specific location.
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def _restore_default_executable()

    @abc.abstractmethod
    def _execute_a_test (self, testcase, exe_path_map, env_vars, \
                    callback_object=None, timeout=None, collect_output=None):
        """ Execute a test given that the executables have been set 
            properly
        """
    #~ def _execute_a_test()

    @abc.abstractmethod
    def _do_generate_tests (self, exe_path_map, code_builds_factory, 
                                        meta_criteria_tool_obj=None, \
                                        max_time=None):
        print ("!!! Must be implemented in child class !!!")
    #~ def _do_generate_tests()
#~ class BaseTestcaseTool


#
# [LICENCE]
#
""" From dict criteria test execution optimizer class
"""

from __future__ import print_function
import os
import sys
import copy
import logging

import muteria.common.mix as common_mix

from muteria.drivers.optimizers.criteriatestexecution.\
                                base_criteria_test_execution_optimizer \
                                    import BaseCriteriaTestExecutionOptimizer

from muteria.drivers.optimizers.testexecution.tools.default \
                                                import TestExecutionOptimizer

from muteria.drivers import DriversUtils

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaTestExecutionOptimizer(BaseCriteriaTestExecutionOptimizer):

    def __init__(self, config, explorer, criterion, dictobj=None):
        ERROR_HANDLER.assert_true(dictobj is not None, \
                                        "dictobj cannot be None", __file__)
        BaseCriteriaTestExecutionOptimizer.__init__(self, config, explorer, \
                                                                    criterion)
        # test objectives and tests are meta here
        self.dictobj = dictobj
    #~ def __init__()

    #######################################################################
    ##################### Methods implemented ############################
    #######################################################################

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        return True
    #~ def installed()

    def reset (self, toolalias, test_objective_list, test_list, **kwargs):
        """ Reset the optimizer
        """
        # Make sure that all objectives in test_objective_list are in dictobj
        # and get test_of_testobjective
        test_of_to = {}
        for to in test_objective_list:
            meta_to = DriversUtils.make_meta_element(to, toolalias)
            ERROR_HANDLER.assert_true(meta_to in self.dictobj, \
                                "meta test objective {} not in self.dictobj"\
                                                    .format(meta_to), __file__)
            test_of_to[to] = self.dictobj[meta_to]

            # If None as testlist, use all tests
            if test_of_to[to] is None:
                test_of_to[to] = test_list

        self.test_objective_ordered_list = copy.deepcopy(test_objective_list)
        self.pointer = 0
        self.test_objective_to_test_execution_optimizer = {
            to: TestExecutionOptimizer(self.config, self.explorer) \
            for to in self.test_objective_ordered_list
        }
        for to, teo in self.test_objective_to_test_execution_optimizer.items():
            teo.reset(None, test_of_to[to], disable_reset=True)
    #~ def reset()
#~ class CriteriaTestExecutionOptimizer




#
# [LICENCE]
#
""" Default criteria test execution optimizer class
"""

from __future__ import print_function
import os
import sys
import copy
import logging

import muteria.common.mix as common_mix
import muteria.common.matrices as common_matrices

import muteria.controller.explorer as explorer
from muteria.drivers.criteria import TestCriteria

import muteria.drivers.optimizers.criteriatestexecution.tools.\
                                        strongmutation_by_weakmutation as sm_wm

from muteria.drivers.optimizers.testexecution.tools.default \
                                                import TestExecutionOptimizer

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaTestExecutionOptimizer(sm_wm.CriteriaTestExecutionOptimizer):

    #######################################################################
    ##################### Methods implemented ############################
    #######################################################################

    ##### Private methods #####
    
    def _get_optimizing_criterion(self):
        return TestCriteria.MUTANT_COVERAGE
    #~ def _get_optimizing_criterion()
#~ class CriteriaTestExecutionOptimizer




#
#> python3 -m pip install --user --upgrade setuptools wheel
#
#> python3 setup.py sdist bdist_wheel
#> python3 -m pip install --user --upgrade twine
#> python3 -m twine upload dist/muteria-<version>.tar.gz
#> rm -rf dist build muteria.egg-info __pycache__/
#

import os
from setuptools import setup, find_packages

from _version import __version__, _framework_name
 
thisdir = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(thisdir, "README.md"), "r") as fh:
    long_description = fh.read()

setup(
    # This is the name of your PyPI-package.
    name=_framework_name,    
    python_requires='>3.3.0',
    # Update the version number for new releases
    version=__version__,
    description='Software Analysis and Testing Framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Thierry Titcheu Chekam',
    author_email='thierry_tct@yahoo.com',
    url='https://github.com/muteria/muteria',
    packages = ['muteria'],
    #packages=find_packages(),
    py_modules = ['_version'],
    include_package_data=True,
    install_requires = [
            "numpy",
            "pandas",
            "scipy",
            "matplotlib",
            "networkx",
            "gitpython",
            "tqdm",
            "joblib",
            "docker", # https://docker-py.readthedocs.io/en/stable/index.html

            # SERVER
            "flask",
            "flask_socketio",
    ],
    # The name of your scipt, and also the command you'll be using for calling it
    #scripts=['cli/muteria'],
    entry_points={
        'console_scripts': [
            'muteria=muteria.cli.cli:main',
        ],
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ]
)




from __future__ import print_function

import sys
import os
import logging
import shutil

import muteria.common.mix as common_mix

import muteria.repositoryandcode.codes_convert_support as ccs

ERROR_HANDLER = common_mix.ErrorHandler

__all__ = ['FromLLVMBitcode']

class FromLLVMBitcode(ccs.BaseCodeFormatConverter):
    def __init__(self):
        self.src_formats = [
            ccs.CodeFormats.LLVM_BITCODE,
        ]
        self.dest_formats = [
            ccs.CodeFormats.LLVM_BITCODE,
            ccs.CodeFormats.OBJECT_FILE,
            ccs.CodeFormats.NATIVE_CODE,
        ]
    #~ def __init__()

    def convert_code(self, src_fmt, dest_fmt, file_src_dest_map, \
                                                        repository_manager):
        ERROR_HANDLER.error_exit("Must Implement", __file__)
    #~ def identity_function()

    def get_source_formats(self):
        return self.src_formats
    #~ def get_source_formats()

    def get_destination_formats_for(self, src_fmt):
        return self.dest_formats
    #~ def get_destination_formats()
#~ class FromLLVMBitcode()


""" This module contains functions to setup the logging facility.
    The logging facility make use of the standard logging lib.

    - The function `console_tmp_log_setup` set the format of the log
        to stdout to display logs before the directory containing the 
        project's log files is created. 
        (allow to pretty print status or error on stdout)
    - The function `setup` sets up the log facility to log into files.
        The function must be called only Once
"""


from __future__ import print_function

import logging
import logging.handlers

# private static variable of this module
_SETUP_DONE = False

def is_setup():
    return _SETUP_DONE
#def is_setup()

def setup(logfile=None, logconsole=False, file_level=logging.INFO, 
            console_level=logging.INFO, file_max_bytes=20000, 
            n_file_backups=1, root_name=''):

    global _SETUP_DONE

    # skip if already setup        
    if _SETUP_DONE:
        return

    # create logger 
    logger = logging.getLogger(root_name)
    logger.setLevel(min(file_level, console_level))

    # create formatter and add it to the handlers
    formatter = logging.Formatter(\
                    fmt='%(asctime)s [%(name)s] [%(levelname)s] %(message)s', \
                    datefmt='%m/%d/%Y %I:%M:%S %p')

    # create file handler?
    if logfile is not None:
        #fh = logging.FileHandler(logfile, mode='w')
        fh = logging.handlers.RotatingFileHandler(logfile, mode='a', \
                        maxBytes=file_max_bytes, backupCount=n_file_backups)
        fh.setLevel(file_level)
        fh.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(fh)

    # create console handler?
    if logconsole:
        ch = logging.StreamHandler()
        ch.setLevel(console_level)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(ch)

    _SETUP_DONE = True
#~ def setup()

def console_tmp_log_setup(loglevel=logging.INFO, root_name=''):
    logging.basicConfig(level=loglevel, \
                format='%(asctime)s [%(name)s] [%(levelname)s] %(message)s', \
                datefmt='%m/%d/%Y %I:%M:%S %p')
#~ def console_tmp_log_setup()



import os
import subprocess
import logging

from muteria.drivers import DriversUtils

from muteria.common.mix import GlobalConstants

def make_build_func(repo_root_dir, exe_rel_paths, compiler, flags_list, clean,\
                                                                reconfigure):
    """ Helper for GNU make
    """

    def print_err(out, msg):
        out = out.splitlines()
        #print(out, msg)
        logging.error(str(out), msg)
    #~ def print_err()

    cwd = os.getcwd()
    os.chdir(repo_root_dir)
    
    #try:
    tmp_env = os.environ.copy()
    if compiler is not None:
        tmp_env["CC"] = compiler
    if flags_list is not None:
        tmp_env["CFLAGS"] = " ".join(flags_list)
    
    if reconfigure:
        if os.path.isfile('configure'):
            args_list = ['configure'] 
            retcode, out, _ = DriversUtils.execute_and_get_retcode_out_err(\
                                    prog='bash', args_list=args_list, \
                                    env=tmp_env, merge_err_to_out=True)
        elif os.path.isfile('CMakeLists.txt'):
            args_list = [] 
            retcode, out, _ = DriversUtils.execute_and_get_retcode_out_err(\
                                    prog='cmake', args_list=args_list, \
                                    env=tmp_env, merge_err_to_out=True)
        else:
            retcode = 0
        if retcode != 0:
            print_err(out, "reconfigure failed")
            os.chdir(cwd)
            return GlobalConstants.COMMAND_FAILURE 
        clean = True
    if clean:
        args_list = ['clean']
        retcode, out, _ = DriversUtils.execute_and_get_retcode_out_err(\
                                prog='make', args_list=args_list, \
                                env=tmp_env, merge_err_to_out=True)
        if retcode != 0:
            print_err(out, "clean failed")
            os.chdir(cwd)
            return GlobalConstants.COMMAND_FAILURE 
    
    retcode, out, _ = DriversUtils.execute_and_get_retcode_out_err(\
                        prog='make', env=tmp_env, merge_err_to_out=True)
    if retcode != 0:
        print_err(out, "make")
        os.chdir(cwd)
        return GlobalConstants.COMMAND_FAILURE 
    #except:
    #    os.chdir(cwd)
    #    assert False, "Build Unexpected Error in "+__file__
        #return GlobalConstants.COMMAND_FAILURE

    os.chdir(cwd)
    return GlobalConstants.COMMAND_SUCCESS
#~ def make_build_func()



#
'''
    fs
    --
    This module implements some basic file system operation that are useful
    for loading and storing data
'''

from __future__ import print_function
import os
import json
import tarfile
import zipfile
import time
import shutil
import logging
import pandas as pd

import muteria.common.mix as common_mix

ERROR_HANDLER = common_mix.ErrorHandler

# JSON
def loadJSON (in_file_pathname):
    '''
    Load a Json from file and return it as a python object.

    :param in_file_pathname: Pathname of the Json file to load.
    :returns: loaded data as corresponding python object.
    '''
    with open(in_file_pathname) as fp:
        return json.load(fp)
#~ loadJSON()

def dumpJSON (data_object, out_file_pathname, pretty=False):
    '''
    Store a data object in Json format into a file.

    :param data_object: data to store in Json format. 
    :param out_file_pathname: Pathname of the Json file to store the data.
    :param pretty: Enables visual friendly layout of json file (spaces).
    :returns: None on success and error message on failure.
    '''
    with open(out_file_pathname, "w") as fp:
        if pretty:
            json.dump(data_object, fp, indent=2, sort_keys=True)
        else:
            json.dump(data_object, fp)

    return None
#~ dumpJSON()         

# CSV
def loadCSV (in_file_pathname, separator=" "):
    '''
    Load a CSV from file and return it as a pandas dataframe.

    :param in_file_pathname: Pathname of the CSV file to load.
    :param separator: The separator used in the CSV file. dafault is space.
    :returns: loaded csv data as pandas dataframe.
    '''
    return pd.read_csv(in_file_pathname, sep=separator, index_col=False)
#~ loadCSV()

def dumpCSV (dataframe, out_file_pathname, separator=" "):
    '''
    Store a dataframe in CSV format into a file.

    :param dataframe: data to store in CSV format. 
    :param out_file_pathname: Pathname of the CSV file to store the dataframe.
    :param separator: Separator used in the CSV file.
    :returns: None on success and error message on failure.
    '''
    dataframe.to_csv(out_file_pathname, sep=separator, index=False)

    return None
#~ dumpCSV()         

class TarGz:
    """
        File preperties preserving archiving (using tar gz)
    """

    archive_ext = ".tar.gz"

    open_read_flag = "r:gz"
    open_write_flag = "w:gz"
    opening_function = tarfile.open

    is_archive_file = tarfile.is_tarfile

    get_element = lambda handle, elem: handle.getmember(elem)

    @classmethod
    def _add_fd_function(cls, handle, in_directory, *args, **kwargs):
        handle.add(in_directory, *args, **kwargs)
    #~ def _add_fd_function()

    #################################################################
    ######################## PUBLIC INTERFACE #######################
    #################################################################

    @classmethod
    def get_archive_filename_of(cls, file_dir):
        return file_dir + cls.archive_ext
    #~ def get_archive_filename_of()

    @classmethod
    def compressDir (cls, in_directory, out_archive_pathname=None, 
                    remove_in_directory=False):

        ERROR_HANDLER.assert_true(os.path.isdir(in_directory), \
                                        "invalid in_directory: "+in_directory)

        err_msg = cls._compressFileOrDir(in_directory, out_archive_pathname)

        if remove_in_directory:
            shutil.rmtree(in_directory)

        return err_msg
    #~ def compressDir()

    @classmethod
    def compressFile (cls, in_file, out_archive_pathname=None, 
                    remove_in_file=False):

        ERROR_HANDLER.assert_true(os.path.isfile(in_file), \
                                        "invalid in_file: "+in_file)

        err_msg = cls._compressFileOrDir(in_file, out_archive_pathname)

        if remove_in_file:
            os.remove(in_file)

        return err_msg
    #~ def compressFile()

    @classmethod
    def _compressFileOrDir (cls, in_directory, out_archive_pathname=None):
        '''
        Compress (Archive) a directory to save up disk space and inodes

        :param in_directory: Directory to compress (archive). 
        :param out_archive_pathname: Optional Pathname of the compressed file. 
            If None, the :param:`in_directory` name is used with extension 
            cls.archive_ext added.
        :param remove_in_directory: Decide whether the compressed directory
            should be deleted after compression
        :returns: None on success and an error message on failure
        '''

        if out_archive_pathname is None:
            out_archive_pathname = in_directory + cls.archive_ext

        with cls.opening_function(out_archive_pathname, cls.open_write_flag) \
                                                                    as handle:
            push_dir = os.getcwd()
            os.chdir(os.path.dirname(os.path.abspath(in_directory)))
            cls._add_fd_function(handle, os.path.basename(in_directory)) #, \
                                                                #arcname='.')
            os.chdir(push_dir)

        if not cls.is_archive_file(out_archive_pathname):
            errmsg = " ".join(["The created", cls.archive_ext, "file", \
                                        out_archive_pathname, "is invalid"])
            return errmsg

        return None
    #~ def _compressFileDir()

    @classmethod
    def decompressDir (cls, in_archive_pathname, out_directory=None, 
                                                    remove_in_archive=False):
        err_msg = cls._decompressFileOrDir(in_archive_pathname, \
                                            out_directory, remove_in_archive)
        return err_msg
    #~ def decompressDir()

    @classmethod
    def decompressFile (cls, in_archive_pathname, out_directory=None, 
                                                    remove_in_archive=False):
        err_msg = cls._decompressFileOrDir(in_archive_pathname, \
                                            out_directory, remove_in_archive)
        return err_msg
    #~ def decompressFile()

    @classmethod
    def _decompressFileOrDir (cls, in_archive_pathname, out_directory=None, 
                                                    remove_in_archive=False):
        '''
        Decompress (UnArchive) a directory's tar file.

        :param in_archive_pathname: Tar file to decompress. 
        :param out_directory: Optional pathname of the destination. 
            If None, the :param:`in_archive_pathname` name is used 
            stripping extension `tar.gz`.
        :param remove_in_archive: Decide whether the decompressed file
            should be deleted after decompression
        :returns: None on success and an error message on failure
        '''
        if (in_archive_pathname.endswith(cls.archive_ext)):

            if out_directory is None:
                out_directory = os.path.dirname(in_archive_pathname)

            ERROR_HANDLER.assert_true(os.path.isdir(out_directory), \
                                    "Extract location is missing", __file__)

            with cls.opening_function(in_archive_pathname, cls.open_read_flag)\
                                                                    as handle:
                handle.extractall(path=out_directory)
            
    #        if not os.path.isdir(out_directory):
    #            errmsg = " ".join(["The out_directory", out_directory, \
    #                                "is missing after decompress"])
    #            return errmsg
    #    elif (in_archive_pathname.endswith(".tar")):
    #        if out_directory is None:
    #            out_directory = in_archive_pathname[:-len('.tar')]
    #        if os.path.isdir(out_directory):
    #            shutil.rmtree(out_directory)
    #        tar = tarfile.open(in_archive_pathname, "r:")
    #        tar.extractall()
    #        tar.close()
        else:
            errmsg = " ".join(["Invalid", cls.archive_ext, "file:", \
                                                        in_archive_pathname])

        if remove_in_archive:
            os.remove(in_archive_pathname)

        return None
    #~ def _decompressFileOrDir()

    @classmethod
    def extractFromArchive (cls, in_archive_pathname, extract_pathname, \
                                        out_location=None, is_folder=False):
        if (in_archive_pathname.endswith(cls.archive_ext)):

            if out_location is None:
                out_location = os.path.dirname(in_archive_pathname)

            with cls.opening_function(in_archive_pathname, cls.open_read_flag)\
                                                                    as handle:
                try:
                    cls.get_element(handle, extract_pathname)
                except KeyError:
                    errmsg = " ".join(["Member", extract_pathname, \
                                    "abscent in archive", in_archive_pathname])
                    return errmsg
                handle.extract(extract_pathname, path=out_location)
            
            dest = os.path.join(out_location, extract_pathname)
            if is_folder:
                if not os.path.isdir(dest):
                    errmsg = " ".join(["The extracted directory", dest, \
                                                "is missing after decompress"])
                    return errmsg
            else:
                if not os.path.isfile(dest):
                    errmsg = " ".join(["The extracted file", dest, \
                                                "is missing after decompress"])
                    return errmsg
                
    #    elif (in_archive_pathname.endswith(".tar")):
    #        if out_directory is None:
    #            out_directory = in_archive_pathname[:-len('.tar')]
    #        if os.path.isdir(out_directory):
    #            shutil.rmtree(out_directory)
    #        tar = tarfile.open(in_archive_pathname, "r:")
    #        tar.extractall()
    #        tar.close()
        else:
            errmsg = " ".join(["Invalid", cls.archive_ext, "file:", \
                                                        in_archive_pathname])

        return None
    #~ def extractFromArchive()

    @classmethod
    def removeFromArchive (cls, in_archive_pathname, rem_pathname, \
                                                            is_folder=False):
        # Decompress, remove and compress
        tmp_dir = in_archive_pathname + ".tmp_dir"
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)
        os.mkdir(tmp_dir)
        errmsg = cls.decompressDir (in_archive_pathname, out_directory=tmp_dir, 
                                                        remove_in_archive=True)
        if errmsg is not None:
            return errmsg + " -- (removeFromArchive)" 
        rem = os.path.join(tmp_dir, rem_pathname)
        if is_folder:
            shutil.rmtree(rem)
        else:
            os.remove(rem)
        errmsg = cls.compressDir(tmp_dir, in_archive_pathname, \
                                                    remove_in_directory=True)
        if errmsg is not None:
            return errmsg + " -- (removeFromArchive)" 
        return None
    #~ def removeFromArchive ()
#~ class TarGz

class Zip (TarGz):
    """
        File properties non preserving but with option to addd to archive
    """

    #######################################################################
    ######################### OVERRIDES ###################################
    #######################################################################

    archive_ext = ".zip"

    open_read_flag = "r"
    open_write_flag = "w"
    opening_function = zipfile.ZipFile

    is_archive_file = zipfile.is_zipfile

    get_element = lambda handle, elem: handle.getinfo(elem)

    @classmethod
    def _add_fd_function(cls, handle, in_directory, *args, **kwargs):
        # setup file paths variable
        file_paths = []
   
        if os.path.isfile(in_directory):
            file_paths.append(in_directory)
        else:
            # Read all directory, subdirectories and file lists
            for root, directories, files in os.walk(in_directory):
                for filename in files:
                    # Create the full filepath by using os module.
                    file_path = os.path.join(root, filename)
                    file_paths.append(file_path)
         
        for file_path in file_paths:
            handle.write(file_path, *args, **kwargs)
    #~ def _add_fd_function()

    #######################################################################
    #######################   ADDITIONAL METHOD   #########################
    #######################################################################

    @classmethod
    def addToArchive (cls, archive_pathname, added_pathname, \
                                    in_archive_name=None, is_folder=False):
        """ Create archive if not existing
        """
        if (archive_pathname.endswith(cls.archive_ext)):

            with cls.opening_function(archive_pathname, "a") as handle:
                if not is_folder:
                    handle.write(added_pathname, arcname=in_archive_name)
                else:
                    cls._add_folder_function(handle, added_pathname) #TODO

                # Check
                try:
                    cls.get_element(handle, in_archive_name)
                except KeyError:
                    errmsg = " ".join(["added element", added_pathname, \
                                    "With archive name", in_archive_name,
                                    "abscent in archive", archive_pathname])
                    return errmsg
        else:
            errmsg = " ".join(["Invalid", cls.archive_ext, "file:", \
                                                            archive_pathname])

        return None
    #~ def addToArchive ()
#~ class Zip

class FileDirStructureHandling(object):
    '''
    Can be used for the organization of the output directory.
    Provides methods to access the files and directories
    Can get, create, remove, get_or_create files and dirs
    '''
    def __init__(self, top_dir, top_dir_key, file_dir_dict):
        self.top_dir = top_dir
        self.top_dir_key = top_dir_key
        self.error_module = ERROR_HANDLER
        self.file_dir_to_path_dict = {self.top_dir_key: '.'}
        for fd in file_dir_dict:
            if type(file_dir_dict[fd]) not in (list, tuple):
                self.error_module.error_exit(err_string="%s %s" % \
                        ("Each value in file_dir_dict", \
                        "must be an ordered list of strings"), \
                            call_location=__file__)
            if len(file_dir_dict[fd]) < 1:
                self.error_module.error_exit(err_string="%s %s" % \
                        ("empty path elements for file/dir:", fd), \
                            call_location=__file__)
            if fd in self.file_dir_to_path_dict:
                self.error_module.error_exit(err_string="%s %s %s" % \
                        ("file or directory already appears in", \
                         "self.file_dir_to_path_dict:", fd), \
                            call_location=__file__)
            # Set the relative path
            self.file_dir_to_path_dict[fd] = os.path.join(*file_dir_dict[fd])


    def resolve(self, name):
        """
        """
        if name not in self.file_dir_to_path_dict:
            self.error_module.error_exit(err_string="%s %s" % \
                                    (name, "not in file_dir_to_path_dict"), \
                                        call_location=__file__ )
        return self.file_dir_to_path_dict[name]

    def get_file_pathname(self, filename, rel_path=False):
        if rel_path:
            return self.resolve(filename)
        return os.path.normpath( \
                            os.path.join(self.top_dir, self.resolve(filename)))

    def get_existing_file_pathname(self, filename, rel_path=False):
        fullpathstring = self.get_file_pathname(filename, rel_path=False)
        if rel_path:
            retpathstring = self.get_file_pathname(filename, rel_path=True)
        else:
            retpathstring = fullpathstring
        if not os.path.isfile(fullpathstring):
            self.error_module.error_exit(err_string="%s %s" % \
                    ("getting a file non existing", fullpathstring), \
                       call_location=__file__)
        return retpathstring

    def file_exists(self, filename):
        fullpathstring = self.get_file_pathname(filename, rel_path=False)
        return os.path.isfile(fullpathstring)
        
    def remove_file_and_get(self, filename, rel_path=False):
        fullpathstring = self.get_file_pathname(filename, rel_path=False)
        if os.path.isfile(fullpathstring):
            os.remove(fullpathstring)
        if rel_path:
            return self.get_file_pathname(filename, rel_path=True)
        return fullpathstring

    def get_dir_pathname(self, dirname, rel_path=False):
        return self.get_file_pathname(dirname, rel_path)

    def dir_exists(self, dirname):
        fullpathstring = self.get_dir_pathname(dirname, rel_path=False)
        return os.path.isdir(fullpathstring)

    def get_existing_dir_pathname(self, dirname, rel_path=False):
        fullpathstring = self.get_dir_pathname(dirname, rel_path=False)
        if rel_path:
            retpathstring = self.get_dir_pathname(dirname, rel_path=True)
        else:
            retpathstring = fullpathstring
        if not os.path.isdir(fullpathstring):
            self.error_module.error_exit(err_string="%s %s" % \
                    ("getting a directory non existing", fullpathstring), \
                        call_location=__file__)
        return retpathstring

    def clean_create_and_get_dir(self, dirname, rel_path=False):
        fullpathstring = self.get_dir_pathname(dirname, rel_path=False)
        if os.path.isdir(fullpathstring):
            shutil.rmtree(fullpathstring)
        os.mkdir(fullpathstring)
        if rel_path:
            return self.get_dir_pathname(dirname, rel_path=True)
        return fullpathstring

    def get_or_create_and_get_dir(self, dirname, rel_path=False):
        fullpathstring = self.get_dir_pathname(dirname, rel_path=False)
        if not os.path.isdir(fullpathstring):
            os.makedirs(fullpathstring)
        if rel_path:
            return self.get_dir_pathname(dirname, rel_path=True)
        return fullpathstring

    def remove_dir_and_get(self, dirname, rel_path=False):
        fullpathstring = self.get_dir_pathname(dirname, rel_path=False)
        if os.path.isdir(fullpathstring):
            shutil.rmtree(fullpathstring)
        if rel_path:
            return self.get_dir_pathname(dirname, rel_path=True)
        return fullpathstring
#~ class FileDirStructureHandling()

class CheckpointState(object):
    EXEC_COMPLETED = "CHECK_POINTED_TASK_COMPLETED"
    EXEC_STARTING = "CHECK_POINTED_TASK_STARTING"
    AGG_TIME_KEY = "AGGREGATED_TIME"
    DETAILED_TIME_KEY = "DETAILED_TIME"
    CHECKPOINT_DATA_KEY = "CHECKPOINT_DATA"

    '''
        The different states are:
            - Destroyed
            - starting
            - Executing
            - Finished (Completed)
    '''
    def __init__(self, store_filepath, backup_filepath):
        self.store_filepath = store_filepath
        self.backup_filepath = backup_filepath
        # make sure that sub task are destroyed, restarted
        # when parent is. (Not necessary for finished)
        self.dep_checkpoint_states = set()
        self.started = False
        self.finished = False

        # time when the current execution started
        self.starttime = None
        
        # aggregated time loaded from last ckeckpoint (offset by starttime)
        self.prev_aggregated_time = None

        raw_obj = self._get_from_file()
        self._update_this_object(raw_obj)
    #~ def __init__()

    def get_dep_checkpoint_states(self):
        return self.dep_checkpoint_states
    #~ def get_dep_checkpoint_states()
    
    def add_dep_checkpoint_state(self, dep_cp):
        self.dep_checkpoint_states.add(dep_cp)
    #~ def add_dep_checkpoint_state()

    def destroy_checkpoint(self):
        for dep_cp in self.dep_checkpoint_states:
            dep_cp.destroy_checkpoint()
        if os.path.isfile(self.backup_filepath):
            os.remove(self.backup_filepath)
        if os.path.isfile(self.store_filepath):
            #shutil.copy2(self.store_filepath, self.backup_filepath)
            os.remove(self.store_filepath)
        self.started = False
        self.finished = False
        self.starttime = None
        self.prev_aggregated_time = None
    #~ def destroy_checkpoint()

    def set_finished(self, detailed_exectime_obj=None):
        if not self.started:
            ERROR_HANDLER.error_exit("%s" % \
                    "finishing checkpointed task while not started", __file__)
            
        self.started = False
        self.finished = True
        self.write_checkpoint(self.EXEC_COMPLETED, \
                                detailed_exectime_obj=detailed_exectime_obj)
        
        # Freeze time.
        # put this last because used in write_checkpoint
        ## update prev_aggregated_time
        self.prev_aggregated_time += (time.time() - self.starttime)
        ## invalidate startime
        self.starttime = None 
    #~ def set_finished()

    def is_destroyed(self):
        no_files = True
        for dep_cp in self.dep_checkpoint_states:
            no_files &= dep_cp.is_destroyed()
        no_files &= not os.path.isfile(self.store_filepath) and \
                    not os.path.isfile(self.backup_filepath)
        return no_files
    #~ def is_destroyed()

    def is_finished(self):
        return self.finished
    #~ def is_finished()

    def restart_task(self):
        for dep_cp in self.dep_checkpoint_states:
            dep_cp.restart_task()
        self.started = True
        self.finished = False
        self.prev_aggregated_time = 0.0
        self.starttime = time.time()
        self.write_checkpoint(self.EXEC_STARTING)
    #~ def restart_task()

    def load_checkpoint_or_start(self, ret_detailed_exectime_obj=False):
        '''
        This function also show a fresh starting of the execution
        Return None as checkpoint data if start (was not yet started)
                If ret_detailed_exectime_obj is enable, return 
                detailed_exectime_obj as second returned value
        Note: In case of continue from a checkpoint checkpoint,
                the object already in sync with files here (see __init__)
        '''
        res = None
        raw_obj = self._get_from_file()
        if raw_obj is None:
            self.restart_task()
        else:
            res = raw_obj[self.CHECKPOINT_DATA_KEY]
            if res in [self.EXEC_STARTING, self.EXEC_COMPLETED]:
                res = None
            if ret_detailed_exectime_obj:
                res = (res, raw_obj[self.DETAILED_TIME_KEY])
        return res 
    #~ def load_checkpoint_or_start()

    def write_checkpoint(self, json_obj, detailed_exectime_obj=None):
        remove_back = False
        if os.path.isfile(self.store_filepath):
            shutil.copy2(self.store_filepath, self.backup_filepath)
        else:
            remove_back = True
        cur_agg_time = \
                    self.prev_aggregated_time + (time.time() - self.starttime)
        raw_obj = { \
                    self.AGG_TIME_KEY: cur_agg_time, \
                    self.DETAILED_TIME_KEY: detailed_exectime_obj, \
                    self.CHECKPOINT_DATA_KEY: json_obj, \
        }
        dumpJSON(raw_obj, self.store_filepath, pretty=True)
        if remove_back and os.path.isfile(self.backup_filepath):
            os.remove(self.backup_filepath)
    #~ def write_checkpoint()

    def get_execution_time(self):
        ERROR_HANDLER.assert_true(not self.is_destroyed(), \
                                "Trying to get time for destroyed checkpoint",
                                                                    __file__)
        if self.starttime is None:
            return self.prev_aggregated_time
        return self.prev_aggregated_time + (time.time() - self.starttime)
    #~ def get_execution_time()

    def get_detailed_execution_time(self):
        ERROR_HANDLER.assert_true(not self.is_destroyed(), \
                        "Trying to get detailed time for destroyed checkpoint",
                                                                    __file__)

        raw_obj = self._get_from_file()
        return raw_obj[self.DETAILED_TIME_KEY]
    #~ def get_detailed_execution_time()

    def _get_from_file(self):
        contain = None
        trybackup = True
        if os.path.isfile(self.store_filepath):
            try:
                contain = loadJSON(self.store_filepath)
                trybackup = False
            except ValueError:
                trybackup = True
        if trybackup and os.path.isfile(self.backup_filepath):
            try:
                contain = loadJSON(self.backup_filepath)
            except ValueError:
                ERROR_HANDLER.error_exit("%s %s" % (\
                                        "Both Checkpoint store_file and", \
                                        "backup file are invalid"), __file__)
            if not common_mix.confirm_execution("%s %s" % ( \
                        "The checkpoint store_file is invalid but backup", \
                        "is valid. Do you want to use backup?")):
                ERROR_HANDLER.error_exit("%s %s" % (\
                                    "Execution terminated due to", \
                                    "invalid Checkpoint store_file"), __file__)
        
        # Check consistency or update obj
        if contain is not None:
            for key in [self.DETAILED_TIME_KEY, self.AGG_TIME_KEY, \
                                                    self.CHECKPOINT_DATA_KEY]:
                if key not in contain:
                    file_used = self.backup_filepath if trybackup \
                                                    else self.store_filepath
                    ERROR_HANDLER.error_exit("%s (%s). %s %s" % \
                                ("Invalid checkpoint file", file_used, \
                                "do not contain the data for", key), __file__)
        return contain
    #~ def _get_from_file()

    def _update_this_object(self, raw_obj):
        if raw_obj is None:
            # Case of Destroyed state
            self.started = False
            self.finished = False
            self.starttime = None
            self.prev_aggregated_time = None
        else:
            agg_time = raw_obj[self.AGG_TIME_KEY]
            checkpoint_data = raw_obj[self.CHECKPOINT_DATA_KEY]
            if checkpoint_data == self.EXEC_STARTING:
                # Starting State
                self.started = True
                self.finished = False
                self.starttime = time.time()
            elif checkpoint_data == self.EXEC_COMPLETED:
                # Finished state
                self.started = False
                self.finished = True
                self.starttime = None
            else:
                self.started = True
                self.finished = False
                self.starttime = time.time()
            self.prev_aggregated_time = float(agg_time)
    #~ def _update_this_object() 
#~ class CheckpointState




from __future__ import print_function

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix

ERROR_HANDLER = common_mix.ErrorHandler

class TestcasesInfoObject(object):
    """ This class represent the test case informations data structure
    """
    DATA_KEY = "DATA"
    SUMMARY_KEY = "SUMMARY"
    CUSTOM_KEY = "CUSTOM"
    def __init__(self):
        self.data = {
            self.DATA_KEY: {},
            self.SUMMARY_KEY: None,
            self.CUSTOM_KEY: None,
        }
    #~ def __init__()

    def load_from_file(self, file_path):
        self.data = common_fs.loadJSON(file_path)
    #~ def load_from_file()

    def write_to_file(self, file_path):
        common_fs.dumpJSON(self.data, file_path, pretty=True)
    #~ def write_to_file()

    def add_test (self, test_name, **kwargs):
        ERROR_HANDLER.assert_true(not self.has_test(test_name), \
                            "Test is already present in this: {}".format(\
                                                    test_name), __file__)
        self.data[self.DATA_KEY][test_name] = kwargs
    #~def add_test ()

    def has_test(self, test_name):
        return test_name in self.data[self.DATA_KEY]
    #~ def has_test():

    def get_tests_list(self):
        return list(self.data[self.DATA_KEY].keys())
    #~ def get_tests_list()

    def update_using(self, toolname, old2new_tests, old_test_info_obj):
        # Update tests
        for old_test_name, new_test_name in list(old2new_tests.items()):
            ERROR_HANDLER.assert_true(\
                        old_test_info_obj.has_test(old_test_name), \
                        "Test not present in old_test_info_obj: {}".format(\
                                                    old_test_name), __file__)
            
            ERROR_HANDLER.assert_true(not self.has_test(new_test_name), \
                            "Test is already present in this: {}".format(\
                                                    new_test_name), __file__)
            self.data[self.DATA_KEY][new_test_name] = \
                        old_test_info_obj.data[self.DATA_KEY][old_test_name]

        # Update Summary
        if self.data[self.SUMMARY_KEY] is None:
            self.data[self.SUMMARY_KEY] = {}
        self.data[self.SUMMARY_KEY][toolname] = \
                        old_test_info_obj.data[self.SUMMARY_KEY]

        # Update Custom
        if self.data[self.CUSTOM_KEY] is None:
            self.data[self.CUSTOM_KEY] = {}
        self.data[self.CUSTOM_KEY][toolname] = \
                        old_test_info_obj.data[self.CUSTOM_KEY]
    #~ def update_using()

    def remove_test(self, test_name):
        ERROR_HANDLER.assert_true(self.has_test(test_name), \
                        "Removing an unexisting test: {}".format(test_name), \
                                                                    __file__)
        del self.data[self.DATA_KEY][test_name]
    #~ def remove_test()

    def get_summary(self):
        return self.data[self.SUMMARY_KEY]
    #~ def get_summary():

    def get_custom(self):
        return self.data[self.CUSTOM_KEY]
    #~ def get_custom():
#~ class TestcasesInfoObject


"""
- A project is a configuration (can be created, modified/updated, deleted)
- It is possible to switch between projects (every project operation needs)
   the project unique id as part of the url
- Each user operation need the user id as part of the url 
   (for nor default value)
"""

from flask import Flask, render_template, request, session, url_for
from werkzeug import secure_filename
from werkzeug.datastructures import MultiDict, ImmutableMultiDict

import argparse

app = Flask(__name__)

@app.route("/")
def homepage():
   return render_template("home.html")

@app.route('/createproject/<projectname>')
def create_project():
   return render_template('createproject.html')

@app.route('/openproject/<projectname>')
def open_project():
   return render_template('configurations.html')

@app.route('/saveproject/<projectname>', methods = ['POST', 'GET'])
def save_config():
   pass #TODO
   # Make sure that the config is sound and save (with backup)
   if request.method == 'POST':
      result = request.form
      f = request.files['file']
      f.save(secure_filename(f.filename))
      result = MultiDict(result)
      result.add(secure_filename(f.filename), 'file uploaded successfully')
      result = ImmutableMultiDict(result)

@app.route('/execute/<projectname>', methods = ['POST', 'GET'])
def execute_with_saved_conf():
   return render_template("progress.html",result = result)

@app.route('/configurations/<projectname>', methods = ['POST', 'GET'])
def execution_configurations(projectname):
   return render_template("configurations.html")

@app.route('/progress/<projectname>', methods = ['POST', 'GET'])
def execution_progress(projectname):
   return render_template("progress.html",result = result)

@app.route('/report/<projectname>', methods = ['POST', 'GET'])
def execution_report(projectname):
   return render_template("report.html",result = result)

@app.route('/info/<projectname>', methods = ['POST', 'GET'])
def execution_info(projectname):
   return render_template("info.html",result = result)

if __name__ == '__main__':
   parser = argparse.ArgumentParser()
   parser.add_argument('--host', default='127.0.0.1', \
                        help="host to use (set to 0.0.0.0 for public access)")
   parser.add_argument('-p', '--port', type=int, default=5000, \
                                                            help="port to use")
   parser.add_argument('-d', '--debug', action="store_true", \
                                       help="enable debug mode for the server")
   args = parser.parse_args()

   # Start the app
   app.run(host=args.host, port=args.port, debug=args.debug)



from __future__ import print_function

import abc

class TestcaseFormatsCodes(dict):
    """ Set each field to the corresponding test class or None if no class
        is implemented.
    """

    KTEST_V1 = None
    KTEST_V2 = None
    GOOGLE_TESTS = None
    JUNIT_TESTS = None

    @classmethod
    def is_testcase_format_class(cls, testcase_format):
        return testcase_format is not None 
    #~ def is_testcase_format_class()
#~ class TestcaseFormatsCodes

class BaseTestcaseFormat(abc.ABC):
    @abc.abstractmethod
    def get_test_suffixes(self):
        """ Return the test files suffixes
        """
        pass
    #~ def get_test_suffixes()

    @abc.abstractmethod
    def test_duplicates(self, test_filename_suffixes, dir_list, file_list):
        """Find duplicate test in given dir_list (having suffix from 
            test_filename_suffixes, Non recursively) and files in file_list
            :return: list of dupliate elements (each duplicate element 
                    is a list of testcases that are duplicate between
                    each others)
        """
        pass
    #~ def test_duplicates()

    @abc.abstractmethod
    def get_supported_convert_into(self):
        pass
    #~ def get_supported_convert_into()

    @abc.abstractmethod
    def convert_into(self, dest_type, src_testfile, dest_testfile):
        """ Convert src_testfile of this type into dest_testfile of dest_type
        """
        pass
    #~ def convert_into()

    @abc.abstractmethod
    def get_supported_convert_from(self):
        pass
    #~ def get_supported_convert_from()

    @abc.abstractmethod
    def convert_from(self, src_type, src_testfile, dest_testfile):
        """ Convert src_testfile of src_type into dest_testfile of this type
        """
        pass
    #~ def convert_from()
#~ class BaseTestcaseFormat()



from __future__ import print_function

import os
import logging
import shutil
import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.drivers.criteria import TestCriteria

ERROR_HANDLER = common_mix.ErrorHandler

# -----------------------------------------------
# Special
TOP_OUTPUT_DIR_KEY = "main_controller_top_output"

# Directories
## CONSTANTS
CRITERIA_WORKDIR = "criteria_workdir"
TESTSCASES_WORKDIR = "testscases_workdir"
RESULTS_DATA_DIR = "RESULTS_DATA"
RESULTS_MATRICES_DIR = "matrices"
RESULTS_TESTEXECUTION_OUTPUTS_DIR = "testexecution_outputs"
RESULTS_STATS_DIR = "statistics"
OTHER_COPIED_RESULTS = "other_copied_results"

CB_FACTORY_WORKDIR = "code_build_workdir"

CONTROLLER_DATA_DIR = "_controller_dat"
CTRL_CHECKPOINT_DIR = "checkpoint_states"
CTRL_LOGS_DIR = "logs"

EXECUTION_TMP_DIR = "execution_tmp"

# Files
## CONSTANTS
SAVED_CONF = "saved_configuration"
EXECUTION_STATE = "execution_state"
EXECUTION_STATE_BAKUP = "execution_state" + ".bak"
EXECUTION_TIMES = "execution_times"
MAIN_LOG_FILE = "ctrl_log.log"

TEST_PASS_FAIL_MATRIX = "PASSFAIL.csv"
CRITERIA_MATRIX = {}
for criterion in TestCriteria:
    CRITERIA_MATRIX[criterion] = criterion.get_str()+".csv"

STATS_MAIN_FILE_HTML = "main_stats.html"
STATS_MAIN_FILE_JSON = "main_stats.json"

TMP_TEST_PASS_FAIL_MATRIX = "tmp_PASSFAIL.csv"
PARTIAL_TMP_TEST_PASS_FAIL_MATRIX = "partial_tmp_PASSFAIL.csv"
TMP_CRITERIA_MATRIX = {}
for criterion in TestCriteria:
    TMP_CRITERIA_MATRIX[criterion] = "tmp_"+criterion.get_str()+".csv"
PARTIAL_TMP_CRITERIA_MATRIX = {}
for criterion in TestCriteria:
    PARTIAL_TMP_CRITERIA_MATRIX[criterion] = \
                                    "partial_tmp_"+criterion.get_str()+".csv"
TMP_SELECTED_TESTS_LIST = "tmp_selected_test.json"
TMP_SELECTED_CRITERIA_OBJECTIVES_LIST = "tmp_selected_criteria_objectives.json"

# OUTPUT Hashed
PROGRAM_TESTEXECUTION_OUTPUT = "program_output.json" 
TMP_PROGRAM_TESTEXECUTION_OUTPUT = "tmp_program_output.json" 
PARTIAL_TMP_PROGRAM_TESTEXECUTION_OUTPUT = "partial_tmp_program_output.json"
CRITERIA_EXECUTION_OUTPUT = {}
for criterion in TestCriteria:
    CRITERIA_EXECUTION_OUTPUT[criterion] = criterion.get_str()+"_output.json"
TMP_CRITERIA_EXECUTION_OUTPUT = {}
for criterion in TestCriteria:
    TMP_CRITERIA_EXECUTION_OUTPUT[criterion] = \
                                    "tmp_"+criterion.get_str()+"_output.json"
PARTIAL_TMP_CRITERIA_EXECUTION_OUTPUT = {}
for criterion in TestCriteria:
    PARTIAL_TMP_CRITERIA_EXECUTION_OUTPUT[criterion] = \
                            "partial_tmp_"+criterion.get_str()+"_output.json"
# ---------------------------------------------------------

def get_outputdir_structure_by_filesdirs():
    '''
    :returns: The structure of the output directory as directely 
                controlled by this controller. 
                The structure is a 'dict' where keys
                are files or folders and values are None for files and
                another dict for folders.
    '''
    # Dirs
    TopExecutionDir = {
        CRITERIA_WORKDIR: [CRITERIA_WORKDIR],
        TESTSCASES_WORKDIR: [TESTSCASES_WORKDIR],
        RESULTS_DATA_DIR: [RESULTS_DATA_DIR],
        RESULTS_MATRICES_DIR: [RESULTS_DATA_DIR, RESULTS_MATRICES_DIR],
        RESULTS_TESTEXECUTION_OUTPUTS_DIR: [RESULTS_DATA_DIR, \
                                            RESULTS_TESTEXECUTION_OUTPUTS_DIR],
        RESULTS_STATS_DIR: [RESULTS_DATA_DIR, RESULTS_STATS_DIR],
        OTHER_COPIED_RESULTS: [RESULTS_DATA_DIR, OTHER_COPIED_RESULTS],
        CB_FACTORY_WORKDIR: [CB_FACTORY_WORKDIR],
        CONTROLLER_DATA_DIR: [CONTROLLER_DATA_DIR],
        CTRL_CHECKPOINT_DIR: [CONTROLLER_DATA_DIR, CTRL_CHECKPOINT_DIR],
        CTRL_LOGS_DIR: [CONTROLLER_DATA_DIR, CTRL_LOGS_DIR],
        EXECUTION_TMP_DIR: [CONTROLLER_DATA_DIR, EXECUTION_TMP_DIR]
    }

    # Files
    TopExecutionDir[SAVED_CONF] = TopExecutionDir[CTRL_CHECKPOINT_DIR] \
                                    + [SAVED_CONF]
    TopExecutionDir[EXECUTION_STATE] = \
                                    TopExecutionDir[CTRL_CHECKPOINT_DIR] \
                                    + [EXECUTION_STATE]
    TopExecutionDir[EXECUTION_STATE_BAKUP] = \
                                    TopExecutionDir[CTRL_CHECKPOINT_DIR] \
                                    + [EXECUTION_STATE_BAKUP]
    TopExecutionDir[EXECUTION_TIMES] = TopExecutionDir[CONTROLLER_DATA_DIR] \
                                        + [EXECUTION_TIMES]
    TopExecutionDir[MAIN_LOG_FILE] = TopExecutionDir[CTRL_LOGS_DIR] \
                                        + [MAIN_LOG_FILE]

    TopExecutionDir[TEST_PASS_FAIL_MATRIX] = \
                TopExecutionDir[RESULTS_MATRICES_DIR] + [TEST_PASS_FAIL_MATRIX]
    TopExecutionDir[TMP_TEST_PASS_FAIL_MATRIX] = \
                                    TopExecutionDir[RESULTS_MATRICES_DIR] \
                                                + [TMP_TEST_PASS_FAIL_MATRIX]
    TopExecutionDir[PARTIAL_TMP_TEST_PASS_FAIL_MATRIX] = \
                                    TopExecutionDir[RESULTS_MATRICES_DIR] \
                                        + [PARTIAL_TMP_TEST_PASS_FAIL_MATRIX]
    for criterion in TestCriteria:
        TopExecutionDir[CRITERIA_MATRIX[criterion]] = \
                                    TopExecutionDir[RESULTS_MATRICES_DIR] \
                                                + [CRITERIA_MATRIX[criterion]]
        TopExecutionDir[TMP_CRITERIA_MATRIX[criterion]] = \
                                    TopExecutionDir[EXECUTION_TMP_DIR] \
                                            + [TMP_CRITERIA_MATRIX[criterion]]
        TopExecutionDir[PARTIAL_TMP_CRITERIA_MATRIX[criterion]] = \
                                    TopExecutionDir[EXECUTION_TMP_DIR] \
                                    + [PARTIAL_TMP_CRITERIA_MATRIX[criterion]]
    # Output
    TopExecutionDir[PROGRAM_TESTEXECUTION_OUTPUT] = \
                    TopExecutionDir[RESULTS_TESTEXECUTION_OUTPUTS_DIR] + \
                                                [PROGRAM_TESTEXECUTION_OUTPUT]
    TopExecutionDir[TMP_PROGRAM_TESTEXECUTION_OUTPUT] = \
                        TopExecutionDir[RESULTS_TESTEXECUTION_OUTPUTS_DIR] + \
                                            [TMP_PROGRAM_TESTEXECUTION_OUTPUT]
    TopExecutionDir[PARTIAL_TMP_PROGRAM_TESTEXECUTION_OUTPUT] = \
                        TopExecutionDir[RESULTS_TESTEXECUTION_OUTPUTS_DIR] + \
                                    [PARTIAL_TMP_PROGRAM_TESTEXECUTION_OUTPUT]
    for criterion in TestCriteria:
        TopExecutionDir[CRITERIA_EXECUTION_OUTPUT[criterion]] = \
                        TopExecutionDir[RESULTS_TESTEXECUTION_OUTPUTS_DIR] + \
                                        [CRITERIA_EXECUTION_OUTPUT[criterion]]
        TopExecutionDir[TMP_CRITERIA_EXECUTION_OUTPUT[criterion]] = \
                                        TopExecutionDir[EXECUTION_TMP_DIR] + \
                                    [TMP_CRITERIA_EXECUTION_OUTPUT[criterion]]
        TopExecutionDir[PARTIAL_TMP_CRITERIA_EXECUTION_OUTPUT[criterion]] = \
                                    TopExecutionDir[EXECUTION_TMP_DIR] + \
                            [PARTIAL_TMP_CRITERIA_EXECUTION_OUTPUT[criterion]]
    
    TopExecutionDir[TMP_SELECTED_TESTS_LIST] = \
                TopExecutionDir[EXECUTION_TMP_DIR] + [TMP_SELECTED_TESTS_LIST]

    TopExecutionDir[TMP_SELECTED_CRITERIA_OBJECTIVES_LIST] = \
                                TopExecutionDir[EXECUTION_TMP_DIR] \
                                    + [TMP_SELECTED_CRITERIA_OBJECTIVES_LIST]

    TopExecutionDir[STATS_MAIN_FILE_HTML] = TopExecutionDir[RESULTS_STATS_DIR]\
                                                    + [STATS_MAIN_FILE_HTML]

    TopExecutionDir[STATS_MAIN_FILE_JSON] = TopExecutionDir[RESULTS_STATS_DIR]\
                                                    + [STATS_MAIN_FILE_JSON]
                                                    
    return TopExecutionDir
#~ def get_outputdir_structure_by_filesdirs():


class Explorer(common_fs.FileDirStructureHandling):
    def __init__(self, outdir):
        common_fs.FileDirStructureHandling.__init__(self, \
            top_dir=os.path.abspath(outdir), \
            top_dir_key=TOP_OUTPUT_DIR_KEY, \
            file_dir_dict=get_outputdir_structure_by_filesdirs())
    #~ def __init__()
#~ class Explorer

class TopExplorer(object):
    """
    TODO: Handle zipping of history and unzipping...
    Maybe need a zipping explorer?
    """
    LATEST_NAME = "latest"
    HISTORY_PREFIX = 'history'
    def __init__(self, root_outdir):
        self.root_outdir = root_outdir
        self.timeline_outdirs = []

        # get timeline_outdirs
        self.timeline_outdirs.append(os.path.join(self.root_outdir, \
                                                            self.LATEST_NAME))
        
        tmp_outdirs = []
        if os.path.isdir(root_outdir):
            for fd in os.listdir(root_outdir):
                if self._is_history_file_dir(fd):
                    tmp_outdirs.append(os.path.join(self.root_outdir, fd))
        
        tmp_outdirs.sort(key=lambda x: self._history_of_outdir(x))
        tmp_histories = [self._history_of_outdir(x) for x in tmp_outdirs]

        fix_missing_history = False
        for ind in range(len(tmp_histories)):
            new_outdir = tmp_outdirs[ind]
            if tmp_histories[ind] != ind + 1:
                if not fix_missing_history:
                    fix_missing_history = common_mix.confirm_execution(\
                                            "Missing history: {}. {}".format(\
                                ind+1, "Do you want to automatically fix it?"))
                    ERROR_HANDLER.assert_true(fix_missing_history, \
                                    "No history should be missing {}".format(\
                                    "when executing. Fix manually."), __file__)
                else:
                    new_outdir = self._change_history(tmp_outdirs[ind], ind+1)
                    shutil.move(tmp_outdirs[ind], new_outdir)
            self.timeline_outdirs.append(new_outdir)

        # Create explorers
        self.explorer_list = self._create_explorers(self.timeline_outdirs)
    #~ def __init__()

    def get_explorer_list(self, history):
        ERROR_HANDLER.assert_true(history >= 0, \
                                    "Invalid History, must be >=0", __file__)
        ERROR_HANDLER.assert_true(history < len(self.explorer_list), \
                        "Invalid History, must be <= max history", __file__)
        return self.explorer_list[history]
    #~ def get_explorer_list()

    def get_latest_explorer(self):
        return self.explorer_list[0]
    #~ def get_latest_explorer()

    def archive_latest(self):
        new_timeline_outdirs = [self._make_history_name(1)]

        # shift everything
        for history in range(1, len(self.timeline_outdirs)):
            new_timeline_outdirs.append(self._change_history(\
                                    self.timeline_outdirs[history], history+1))

        # moves
        for i in range(len(new_timeline_outdirs)-1, -1, -1):
            shutil.move(self.timeline_outdirs[i], new_timeline_outdirs[i])

        self.timeline_outdirs[1:] = new_timeline_outdirs

        # Re-create explorers
        self.explorer_list = self._create_explorers(self.timeline_outdirs)
    #~ def archive_latest()

    def _create_explorers(self, outdir_list):
        explorer_list = []
        for outdir in outdir_list:
            ERROR_HANDLER.assert_true(outdir.startswith(self.root_outdir),\
                        "outdir {} not in rootdir".format(outdir), __file__)
            explorer_list.append(Explorer(outdir))
        return explorer_list
    #~ def _create_explorers()

    def _make_history_name(self, history, relative=False):
        rel_name = self.HISTORY_PREFIX + "_" + str(history)
        if relative:
            return rel_name
        else:
            return os.path.join(self.root_outdir, rel_name)
    #~ def _make_history_name()

    def _change_history(self, pathname, new_history):
        ERROR_HANDLER.assert_true(self._is_history_file_dir(pathname))
        dir_name, base_name = os.path.split(pathname)
        new_base_name = base_name.replace(\
                                    str(self._history_of_outdir(pathname)), \
                                                        str(new_history), 1)
        new_path_name = os.path.join(dir_name, new_base_name)
        return new_path_name
    #~ def _change_history()

    def _history_of_outdir(self, pathname):
        return int(os.path.basename(pathname).split('_')[1].split('.')[0])
    #~ def _history_of_outdir():

    def _is_history_file_dir(self, pathname):
        parts = os.path.basename(pathname).split('_')
        if len(parts) != 2:
            return False
        if parts[0] != self.HISTORY_PREFIX:
            return False
        if not parts[1].split('.')[0].isdigit():
            return False
        return True
    #~ def _is_history_file_dir()
#~ class TopExplorer



def _make_available():
    import os
    import sys
    import importlib

    import muteria.common.mix as common_mix
    ERROR_HANDLER = common_mix.ErrorHandler
    
    pkg_init_py = os.path.abspath(__file__)
    pkg_path, init_file = os.path.split(pkg_init_py)
    modules_list = []
    # make this pkg visible
    sys.path.insert(0, pkg_path)
    for m in os.listdir(pkg_path):
        # Consider only '.py' files
        if not m.endswith('.py'):
            continue
            
        m_path = os.path.join(os.path.join(pkg_path, m))
        if os.path.isfile(m_path) and m != init_file:
            mod_name = os.path.splitext(m)[0]
            ERROR_HANDLER.assert_true(mod_name not in globals(), \
                    "Invalid module name ({}), change name.".format(mod_name),\
                    __file__)
            ERROR_HANDLER.assert_true('.' not in mod_name, "{} {} {}".format(\
                            "Module name(", mod_name, 
                            ") must not contain dot(.) character"), __file__)
            mod = importlib.import_module(mod_name)

            # Set visible to this package
            globals()[mod_name] = mod
    sys.path.pop(0)
#~ def _make_available()

# Make the different modules visible
_make_available()
del _make_available